'''
modified version of what could be found at:
https://github.com/kbat/mc-tools

This module deals with the parsing of MCNP MCTAL files

GNU Lesser General Public License v3.0
Permissions of this copyleft license are conditioned on making available
complete source code of licensed works and modifications under the same license
or the GNU GPLv3. Copyright and license notices must be preserved.
Contributors provide an express grant of patent rights. However, a larger work
using the licensed work through interfaces provided by the licensed work may be
distributed under different terms and without source code for the larger work.
'''

"""
Copyright 2019 F4E | European Joint Undertaking for ITER and the Development of
Fusion Energy (‘Fusion for Energy’). Licensed under the EUPL, Version 1.2 or - 
as soon they will be approved by the European Commission - subsequent versions
of the EUPL (the “Licence”). You may not use this work except in compliance
with the Licence. You may obtain a copy of the Licence at: 
    https://eupl.eu/1.2/en/  
Unless required by applicable law or agreed to in writing, software distributed
under the Licence is distributed on an “AS IS” basis, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the Licence permissions
and limitations under the Licence.
"""

import sys
import os
import math
import numpy as np
import logging
import pandas as pd


class Header:

    def __init__(self):
        """Header class. Contains a bunch of general information on the mctal
        file
        """
        self.kod = ""  # Name of the code
        self.ver = ""  # Code version
        self.probid = np.array((), dtype=str)  # Date and time when the problem was run
        self.knod = 0                           # The dump number
        self.nps = 0                            # Number of histories that were run
        self.rnr = 0                            # Number of pseudoradom numbers that were used
        self.title = ""                         # Problem identification line
        self.ntal = 0                           # Number of tallies
        self.ntals = np.array((), dtype=int)    # Array of tally numbers
        self.npert = 0                          # Number of perturbations


class Tally:
    """This class is aimed to store all the information contained in a
           tally.
    """

    def __init__(self, tN: int) -> None:
        self.tallyNumber = tN                           # Tally number
        self.typeNumber = 0                             # Particle type number
        self.detectorType = None                        # The type of detector tally where 0=none, 1=point, 2=ring, 3=pinhole radiograph,
                                                        #     4=transmitted image radiograph (rectangular grid),
                                                        #     5=transmitted image radiograph (cylindrical grid)
                                                        # When negative, it provides the type of mesh tally
        self.radiograph = False                         # Flag set to True is the tally is a radiograph tally.
        self.tallyParticles = np.array((), dtype=int)   # List of 0/1 entries indicating which particle types are used by the tally
        self.tallyComment = np.array((), dtype=str)     # The FC card lines
        self.nCells = 0                                 # Number of cell, surface or detector bins
        self.mesh = False                               # True if the tally is a mesh tally
        self.meshInfo = np.array([0, 1, 1, 1], dtype=int)  # Mesh binning information in the case of a mesh tally
        self.nDir = 0                                   # Number of total vs. direct or flagged vs. unflagged bins
        self.nUsr = 0                                   # Number of user bins
        self.usrTC = None                               # Total / cumulative bin in the user bins
        self.nSeg = 0                                   # Number of segment bins
        self.segTC = None                               # Total / cumulative bin in the segment bins
        self.nMul = 0                                   # Number of multiplier bins
        self.mulTC = None                               # Total / cumulative bin in the multiplier bins
        self.nCos = 0                                   # Number of cosine bins
        self.cosTC = None                               # Total / cumulative bin in the cosine bins
        self.cosFlag = 0                                # The integer flag of cosine bins
        self.nErg = 0                                   # Number of energy bins
        self.ergTC = None                               # Total / cumulative bin in the energy bins
        self.ergFlag = 0                                # The integer flag of energy bins
        self.nTim = 0                                   # Number of time bins
        self.timTC = None                               # Total / cumulative bin in the time bins
        self.timFlag = 0                                # The integer flag of time bins

        self.cells = np.array(())                       # Array of cell     bin boundaries
        self.usr = np.array(())                         # Array of user     bin boundaries
        self.seg = np.array(())                         # Array of segments bin boundaries
        self.cos = np.array(())                         # Array of cosine   bin boundaries
        self.erg = np.array(())                         # Array of energy   bin boundaries
        self.tim = np.array(())                         # Array of time     bin boundaries
        self.cora = np.array(())                        # Array of cora     bin boundaries for mesh tallies (or lattices)
        self.corb = np.array(())                        # Array of corb     bin boundaries for mesh tallies (or lattices)
        self.corc = np.array(())                        # Array of corc     bin boundaries for mesh tallies (or lattices)

        self.tfc_jtf = np.array(())                     # List of numbers in the tfc line
        self.tfc_dat = []                               # Tally fluctuation chart data (NPS, tally, error, figure of merit)

        self.detectorTypeList = {
                -6: "smesh", -5: "cmesh", -4: "rmesh",
                # The line below duplicates the line above with short names for
                # tally naming during conversion.
                # See the function getDetectorType to see how this information
                # is used
                -3: "Spherical mesh tally", -2: "Cylindrical mesh tally",
                -1: "Rectangular mesh tally",
                0: "None",  1: "Point",  2: "Ring",
                3: "Pinhole radiograph",
                4: "Transmitted image rdiograph (rectangular grid)",
                5: "Transmitted image radiograph (cylindrical grid)",
                # The line below duplicates the line above, with short names
                # for tally naming during conversion.
                # See the function getDetectorType to see how this information
                # is used
                6: "pi",  7: "tir",  8: "tic"}

        self.particleListShort = {
                1: "Neutron", 2: "Photon", 3: "Neutron + Photon",
                4: "Electron", 5: "Neutron + Electron",
                6: "Photon + Electron", 7: "Neutron + Photon + Electron"}

        self.particleList = (
                #   1         2          3          4      5
                "Neutron", "Photon", "Electron", "Muon", "Tau",
                #         6                 7
                "Electron Neutrino", "Muon Neutrino",
                #     8            9          10         11         12
                "Tau Neutrino", "Proton", "Lambda 0", "Sigma +", "Sigma -",
                #   13           14          15          16
                "Cascade 0", "Cascade -", "Omega -", "Lambda c +",
                #   17              18              19         20
                "Cascade c +", "Cascade c 0", "Lambda b 0", "Pion +",
                #   21             22           23        24
                "Neutral Pion", "Kaon +", "K0 Short", "K0 Long",
                #  25   26     27       28     29      30        31
                "D +", "D 0", "D s +", "B +", "B 0", "B s 0", "Deuteron",
                #  32      33       34              35
                "Triton", "He3", "He4 (Alpha)", "Heavy ions")

        self.binIndexList = ("f", "d", "u", "s", "m", "c", "e", "t", "i", "j",
                             "k")

        self.isInitialized = False
        self.valsErrors = None       # Array of values and errors

    def _initializeValuesVectors(self) -> None:
        """This function initializes the 9-D matrix for the storage of
        values and errors."""

        nCells = self._getNbins("f")
        nCora = self._getNbins("i")
        nCorb = self._getNbins("j")
        nCorc = self._getNbins("k")
        nDir = self._getNbins("d")
        nUsr = self._getNbins("u")
        nSeg = self._getNbins("s")
        nMul = self._getNbins("m")
        nCos = self._getNbins("c")
        nErg = self._getNbins("e")
        nTim = self._getNbins("t")

        self.valsErrors = np.empty(
                (nCells, nDir, nUsr, nSeg, nMul, nCos, nErg, nTim,
                 nCora, nCorb, nCorc, 2), dtype=float)

        self.isInitialized = True

#     def print(self, option=[]):
#         """Tally printer
#         """
#         print("\033[1m[TALLY]\033[0m")
#         print("Tally Number: %5d" % self.tallyNumber)

#         print("Tally comment(s):")
#         for comment in self.tallyComment:
#             print("\t%s" % comment)

#         mt = "Yes" if self.mesh else "No"
#         print("Mesh tally: %s" % mt)
#         rt = "Yes" if self.radiograph else "No"
#         print("Radiograph Tally: %s" % rt)
#         print("Detector type: %s" % self.getDetectorType())

#         print("List of particles in tally:")
#         for i, name in enumerate(self.getTallyParticles()):
#             print(r"\t%2d - %s" % (i+1, name))
#         if not self.mesh:
#             print("Number of cells/point detectors/surfaces/macrobodies: %5d" % self.getNbins("f"))
#         else:
#                 mesh_tot = self.getNbins("i")*self.getNbins("j")*self.getNbins("k")

#                 print ("Number of mesh tally bins: %5d" % mesh_tot)
#                 print ("\tNumber of CORA bins: %5d" % self.getNbins("i"))
#                 print ("\tNumber of CORB bins: %5d" % self.getNbins("j"))
#                 print ("\tNumber of CORC bins: %5d" % self.getNbins("k"))
#         print ("Number of tot vs. dir or flag vs. unflag bins: %5d" % self.getNbins("d"))
#         print ("Number of user bins: %5d" % self.getNbins("u"))
#         print ("Number of segments: %5d" % self.getNbins("s"))
#         print ("Number of multipliers: %5d" % self.getNbins("m"))
#         print ("Number of cosine bins: %5d" % self.getNbins("c"))
#         print ("Number of energy bins: %5d" % self.getNbins("e"))
#         print ("Number of time bins: %5d" % self.getNbins("t"))

#         print ("Total values in the tally: %8d" % self.getTotNumber(False))

# Not called
    # def getDetectorType(self, short=False) -> str:
    #     """Returns the type of the detector type used in the tally."""

    #     if not short:
    #         return self.detectorTypeList[self.detectorType]
    #     elif short and self.radiograph:
    #         return self.detectorTypeList[self.detectorType + 3]
    #     elif short and self.mesh:
    #         return self.detectorTypeList[self.detectorType - 3]

# not called anywhere
    # def getTallyParticles(self) -> list[str]:
    #     """Returns the particles used in the tally.
    #        References can be found in Table 4-1 and page B-2 of the
    #        MCNPX manual."""

    #     particleNames = []

    #     if self.typeNumber > 0:
    #         particleNames.append(self.particleListShort[self.typeNumber])
    #     else:
    #         for i, _ in enumerate(self.particleList):
    #             try:
    #                 if self.tallyParticles[i] == 1:
    #                     particleNames.append(self.particleList[i])
    #             except IndexError:
    #                 # For some reasons there can be less than 35 particles
    #                 # listed. Skip in case.
    #                 pass
    #     return particleNames
# This is not called anywhere
    # def getTotNumber(self, includeTotalBin=True) -> int:
    #     """Return the total number of bins."""

    #     nCells = self._getNbins("f", includeTotalBin)
    #     nCora = self._getNbins("i", includeTotalBin)
    #     nCorb = self._getNbins("j", includeTotalBin)
    #     nCorc = self._getNbins("k", includeTotalBin)
    #     nDir = self._getNbins("d", includeTotalBin)
    #     nUsr = self._getNbins("u", includeTotalBin)
    #     nSeg = self._getNbins("s", includeTotalBin)
    #     nMul = self._getNbins("m", includeTotalBin)
    #     nCos = self._getNbins("c", includeTotalBin)
    #     nErg = self._getNbins("e", includeTotalBin)
    #     nTim = self._getNbins("t", includeTotalBin)

    #     tot = (nCells * nDir * nUsr * nSeg * nMul * nCos * nErg * nTim *
    #            nCora * nCorb * nCorc)

    #     return tot

    def _insertCell(self, cN: int) -> bool:
        """Insert cell number."""

        if len(self.cells) <= self.nCells:
            self.cells = np.append(self.cells, cN)
            return True
        else:
            return False

    def _insertCorBin(self, axis: str, value) -> bool:
        """Insert cora/b/c values."""
        if axis == 'a':
            if len(self.cora) <= self.meshInfo[1]+1:
                self.cora = np.append(self.cora, value)
                return True
            else:
                return False
        if axis == 'b':
            if len(self.corb) <= self.meshInfo[2]+1:
                self.corb = np.append(self.corb, value)
                return True
            else:
                return False

        if axis == 'c':
            if len(self.corc) <= self.meshInfo[3]+1:
                self.corc = np.append(self.corc, value)
                return True
            else:
                return False

    def _insertUsr(self, uB: int) -> bool:
        """Insert usr bins."""

        if len(self.usr) <= self.nUsr:
            self.usr = np.append(self.usr, uB)
            return True
        else:
            return False

    def _insertSeg(self, sB: int) -> bool:
        """Insert seg bins."""

        if len(self.seg) <= self.nSeg:
            self.seg = np.append(self.seg, sB)
            return True
        else:
            return False

    def _insertCos(self, cB: int) -> bool:
        """Insert cosine bin."""

        if len(self.cos) <= self.nCos:
            self.cos = np.append(self.cos, cB)
            return True
        else:
            return False

    def _insertRadiograph(self, axis: str, rB: int) -> bool:
        """Insert radiograph coordinates on s and t-axis."""

        if axis == "s":
            if len(self.seg) <= self.nSeg+1:
                self.seg = np.append(self.seg, rB)
                return True
            else:
                return False

        if axis == "t":
            if len(self.cos) <= self.nCos+1:
                self.cos = np.append(self.cos, rB)
                return True
        else:
            return False

    def _insertErg(self, eB: int) -> bool:
        """Insert energy bin."""

        if len(self.erg) <= self.nErg:
            self.erg = np.append(self.erg, eB)
            return True
        else:
            return False

    def _insertTim(self, tB: int) -> bool:
        """Insert time bin."""

        if len(self.tim) <= self.nTim:
            self.tim = np.append(self.tim, tB)
            return True
        else:
            return False

    def _insertTfcJtf(self, jtf: int) -> bool:
        """Insert TFC jtf list."""

        if len(jtf) == 9:
            self.tfc_jtf = jtf
            return True
        else:
            return False

    def _insertTfcDat(self, dat) -> bool:
        """Insert TFC values."""

        if len(dat) <= 4:
            self.tfc_dat.append(dat)
            return True
        else:
            return False

    def _insertValue(self, c, d, u, s, m, a, e, t, f, i, j, k, val) -> None:
        """Insert tally value.
        """
        if self.isInitialized is False:
            self.initializeValuesVectors()

        self.valsErrors[c][d][u][s][m][a][e][t][f][i][j][k] = val

    def _getValue(self, f, d, u, s, m, c, e, t, i, j, k, v) -> float:
        """Return a value from tally.
        """
        return self.valsErrors[f][d][u][s][m][c][e][t][i][j][k][v]

# _getAxis is not called anywhere, check with Davide
    # def _getAxis(self, axis: str):
    #     """Return an array containing the values of the axis bins.
    #     The desired axis is set by passing the
    #     corresponding letter as a function argument as
    #     defined in MCNPX manual (u,s,c,e,t) for the
    #     standard and (i,j,k) for mesh tallies axes (namely
    #     cora/b/c).
    #     """

    #     if axis == "u":
    #         if len(self.usr) != 0:
    #             return np.append([0], self.usr)

    #     if axis == "s":
    #         if len(self.seg) != 0:
    #             if self.radiograph:
    #                 return self.seg
    #             else:
    #                 first = self.seg[0] - 1.
    #                 return np.append([first], self.seg)

    #     if axis == "c":
    #         if len(self.cos) != 0:
    #             if self.radiograph:
    #                 return self.cos
    #             else:
    #                 first = -1.
    #                 return np.append([first], self.cos)

    #     if axis == "e":
    #         if len(self.erg) != 0:
    #             first = 0.0  # self.erg[0] - 1.
    #             return np.append([first], self.erg)

    #     if axis == "t":
    #         if len(self.tim) != 0:
    #             first = self.tim[0] - 1.
    #             return np.append([first], self.tim)

    #     if axis == "i":
    #         return self.cora

    #     if axis == "j":
    #         return self.corb

    #     if axis == "k":
    #         return self.corc

    #     return []

    def _getNbins(self, axis, inclTotBin=True) -> int:
        """Returns the number of bins relative to the desired axis.
        The correspondence is, as usual, (f,d,u,s,m,c,e,t) for standard 8D
        data, plus (i,j,k) for mesh tallies."""

        if axis == "f":
            nCells = 1 if self.nCells == 0 else self.nCells
            return nCells

        if axis == "i":
            return self.meshInfo[1]

        if axis == "j":
            return self.meshInfo[2]

        if axis == "k":
            return self.meshInfo[3]

        if axis == "d":
            nDir = 1 if self.nDir == 0 else self.nDir
            return nDir

        if axis == "u":
            nUsr = 1 if self.nUsr == 0 else self.nUsr
            nUsr = nUsr - 1 if self.usrTC == "t" and not inclTotBin else nUsr
            return nUsr

        if axis == "s":
            nSeg = 1 if self.nSeg == 0 else self.nSeg
            nSeg = nSeg - 1 if self.segTC == "t" and not inclTotBin else nSeg
            return nSeg

        if axis == "m":
            nMul = 1 if self.nMul == 0 else self.nMul
            nMul = nMul - 1 if self.mulTC == "t" and not inclTotBin else nMul
            return nMul

        if axis == "c":
            nCos = 1 if self.nCos == 0 else self.nCos
            nCos = nCos - 1 if self.cosTC == "t" and not inclTotBin else nCos
            return nCos

        if axis == "e":
            nErg = 1 if self.nErg == 0 else self.nErg
            nErg = nErg - 1 if self.ergTC == "t" and not inclTotBin else nErg
            return nErg

        if axis == "t":
            nTim = 1 if self.nTim == 0 else self.nTim
            nTim = nTim - 1 if self.timTC == "t" and not inclTotBin else nTim
            return nTim


class Mctal:
    def __init__(self, filepath: os.PathLike) -> None:
        self.tallies = []
        self.thereAreNaNs = False
        self.header = Header()
        self.mctalFileName = os.path.basename(filepath)
        self.mctalFile = open(filepath, "r")

        # This variable will contain the read lines one by one, but it is
        # important to keep it global because the last value from getHeaders()
        # must be already available as first value for parseTally(). This will
        # also apply to successive calls to parseTally().
        self.line = None
        self.tallies = self._read()
        self.tallydata, self.totalbin = self._get_dfs()

    def get_error_summary(self, include_abs_err: bool = False) -> pd.DataFrame:
        """Return a dataframe containing a summary of the min and max errror
        registered in each tally. If both value and error are equal to zero,
        the errors will be set to NaN, since it means that nothing has been
        scored in the tally.

        Parameters
        ----------
        include_abs_err : bool, optional
            if True includes the absolute error in addition to the total one,
            by default is False

        Returns
        -------
        pd.DataFrame
            error summary
        """
        rows = []
        for tally, data in self.tallydata.items():

            min_error = data['Error'].min()
            min_idx = data['Error'].idxmin()
            max_error = data["Error"].max()
            max_idx = data['Error'].idxmax()

            min_val = data['Value'].iloc[min_idx]
            max_val = data['Value'].iloc[max_idx]

            # put a NaN, since no particle was tallied in the cell
            if min_error == 0 and min_val == 0:
                min_error = np.nan
            if max_error == 0 and max_val == 0:
                max_error = np.nan

            if include_abs_err:
                rows.append([tally, min_error, min_val,
                             max_error, max_val])
            else:
                rows.append([tally, min_error, max_error])

        if include_abs_err:
            columns = ['tally num', 'min rel error', 'min abs err',
                       'max rel error', 'max abs err']
        else:
            columns = ['tally num', 'min rel error', 'max rel error']

        df = pd.DataFrame(rows)
        df.columns = columns

        return df

    def _read(self) -> list[Tally]:
        """This function calls the functions getHeaders and parseTally in
        order to read the entier MCTAL file."""

        logging.info(
              "\n\033[1;34m[Parsing file: %s...]\033[0m" % self.mctalFileName)

        self._getHeaders()
        self._getTallies()

        if self.thereAreNaNs:
            logging.info(
                  "\n \033[1;30mThe MCTAL file contains one or more tallies with NaN values. Flagged.\033[0m\n", file=sys.stderr)

        self.mctalFile.close()

        return self.tallies

    def _getHeaders(self) -> None:
        """This function reads the first lines from the MCTAL file. We call
        "header" what is written from the beginning to the first "tally"
        keyword."""

        self.line = self.mctalFile.readline().split()

        if len(self.line) == 7:
            self.header.kod = self.line[0]
            self.header.ver = self.line[1]
            pID_date = self.line[2]
            self.header.probid = np.append(self.header.probid, pID_date)
            pID_time = self.line[3]
            self.header.probid = np.append(self.header.probid, pID_time)
            self.header.knod = int(self.line[4])
            self.header.nps = int(self.line[5])
            self.header.rnr = int(self.line[6])

        elif len(self.line) == 3:
            self.header.knod = int(self.line[0])
            self.header.nps = int(self.line[1])
            self.header.rnr = int(self.line[2])

        self.header.title = self.mctalFile.readline().strip()

        self.line = self.mctalFile.readline().split()

        self.header.ntal = int(self.line[1])

        if self.header.ntal == 0:
            logging.error(
                  "\n \033[1;31mNo tallies in this MCTAL file. Exiting.\033[0m\n", file=sys.stderr)
            sys.exit(1)

        if len(self.line) == 4:
            self.header.npert = int(self.line[3])
            logging.error(
                  "\n \033[1;31mMCTAL file with perturbation card. Not supported. Exiting.\033[0m\n", file=sys.stderr)
            sys.exit(1)

        self.line = self.mctalFile.readline().split()

        while self.line[0].lower() != "tally":
            for char in self.line:
                self.header.ntals = np.append(self.header.ntals, int(char))
            self.line = self.mctalFile.readline().split()

    def _getTallies(self) -> None:
        """This function supervises the calls to parseTally() function.
        It will keep track of the position of cursor in the MCTAL file and stop execution when EOF is reached."""

        EOF = False
        while not EOF:
            EOF = self._parseTally()

    def _parseTally(self) -> bool:
                """This function parses an entire tally."""
                # TODO the linting is horrible here

                # The first line processed by this function is already in memory, either coming from the
                # last readline() in Header class or from the previous call to parseTally()

                tally = Tally(int(self.line[1]))

                logging.info(
                     " \033[33mParsing tally: %5d\033[0m" % (tally.tallyNumber))

                tally.typeNumber = int(self.line[2])
                if self.line[3] != 0: tally.detectorType = int(self.line[3])

                if tally.detectorType is not None: # check for None is needed for MCNP6 and F1 tally
                        if tally.detectorType >=  3:
                                tally.radiograph = True
                        elif tally.detectorType <= -1:
                                tally.mesh = True

                self.line = self.mctalFile.readline()
                # Some MCTAL files do not have the particle list
                # (e.g. produced by MCNP5) so we need to check if it's
                # present (line with particle list starts with space):
                if self.line[0] == " " and self.line[1] != " ":
                        line = self.line.split()
                        for p in line:
                                if p.isdigit():
                                        v = int(p)
                                        tally.tallyParticles = np.append(tally.tallyParticles, v)
                                else:
                                        print("\n \033[1;31m Problem with particle list in tally %d in %s -> exiting.\033[0m\n" % (tally.tallyNumber, self.mctalFileName), file=sys.stderr)
                                        print(self.line)
                                        sys.exit(1)

                if len(tally.tallyParticles) != 0:
                        self.line = self.mctalFile.readline()

                while self.line[0:5] == " "*5 and self.line[0].lower() != "f":
                        tally.tallyComment = np.append(tally.tallyComment, self.line[5:].rstrip())
                        self.line = self.mctalFile.readline()

                line = self.line.split()

                if not tally.mesh:
                        tally.nCells = int(line[1])
                else:
                        tally.nCells = 1
                        tally.meshInfo[0] = int(line[2]) # Unknown number
                        tally.meshInfo[1] = int(line[3]) # number of cora bins
                        tally.meshInfo[2] = int(line[4]) # number of corb bins
                        tally.meshInfo[3] = int(line[5]) # number of corc bins
                
                # Fix for detector tallies 
                if str(tally.tallyNumber)[-1] == '5':
                        for k in range(tally.nCells):
                                tally.cells = np.append(tally.cells, k + 1)

                self.line = self.mctalFile.readline()

                i = 0
                axisNumber = 0
                axisName = ('a','b','c')
                corsVals = (tally.meshInfo[1], tally.meshInfo[2], tally.meshInfo[3])

                while self.line[0].lower() != "d": # CELLS
                        if tally.mesh:
                                for c in self.line.split():
                                        if not tally._insertCorBin(axisName[axisNumber],float(c)):
                                                raise IOError("Too many cells in the tally n. %d of %s" % (tally.tallyNumber, self.mctalFileName))
                                        i = i + 1
                                        if i == (corsVals[axisNumber] + 1):
                                                axisNumber = axisNumber + 1
                                                i = 0
                        elif "." in self.line and "E" not in self.line:
                                for c in self.line.split():
                                        if not tally._insertCell(float(c)):
                                                raise IOError("Too many cells in the tally n. %d of %s" % (tally.tallyNumber, self.mctalFileName))
                        else:
                                for c in self.line.split():
                                        if not tally._insertCell(int(c)):  # This means that for some reason you are trying to
                                                                          # insert more cells than the number stated in f
                                                raise IOError("Too many cells in the tally n. %d of %s" % (tally.tallyNumber, self.mctalFileName))

                        self.line = self.mctalFile.readline()
                
                if tally.mesh is True:
                    tally.cora = tally.cora[:-1]
                    tally.corb = tally.corb[:-1] 
                    tally.corc = tally.corc[:-1] 
                 
                # DIR
                self.line = self.line.split()
                tally.nDir = int(self.line[1])

                while self.line[0].lower() != "u":
                        self.line = self.mctalFile.readline()

                # USR
                self.line = self.line.split()
                if self.line[0].lower() == "ut": tally.usrTC = "t"
                if self.line[0].lower() == "uc": tally.usrTC = "c"
                tally.nUsr = int(self.line[1])

                # USR BINS
                self.line = self.mctalFile.readline()
                while self.line[0].lower() != "s":
                        for u in self.line.split():
                                if not tally._insertUsr(float(u)):
                                        raise IOError("Too many user bins in the tally n. %d of %s" % (tally.tallyNumber, self.mctalFileName))
                        self.line = self.mctalFile.readline()

                # SEG
                self.line = self.line.split()
                if self.line[0].lower() == "st": tally.segTC = "t"
                if self.line[0].lower() == "sc": tally.segTC = "c"
                tally.nSeg = int(self.line[1])

                # SEGMENT BINS
                self.line = self.mctalFile.readline()
                while self.line[0].lower() != "m":
                        for s in self.line.split():
                                if tally.radiograph:
                                        if not tally._insertRadiograph("s",float(s)):
                                                raise IOError("Too many segment bins in the tally n. %d of %s" % (tally.tallyNumber, self.mctalFileName))
                                else:
                                        if not tally._insertSeg(float(s)):
                                                raise IOError("Too many segment bins in the tally n. %d of %s" % (tally.tallyNumber, self.mctalFileName))
                        self.line = self.mctalFile.readline()

                # MUL
                self.line = self.line.split()
                if self.line[0].lower() == "mt": tally.mulTC = "t"
                if self.line[0].lower() == "mc": tally.mulTC = "c"
                tally.nMul = int(self.line[1])

                while self.line[0].lower() != "c":
                        self.line = self.mctalFile.readline()

                # COS
                self.line = self.line.split()
                if self.line[0].lower() == "ct": tally.cosTC = "t"
                if self.line[0].lower() == "cc": tally.cosTC = "c"
                tally.nCos = int(self.line[1])
                if len(self.line) == 3: tally.cosFlag = int(self.line[2])

                # COSINE BINS
                self.line = self.mctalFile.readline()
                while self.line[0].lower() != "e":
                        for c in self.line.split():
                                if tally.radiograph:
                                        if not tally._insertRadiograph("t",float(c)):
                                                raise IOError("Too many cosine bins in the tally n. %d of %s" % (tally.tallyNumber, self.mctalFileName))
                                else:
                                        if not tally._insertCos(float(c)):
                                                raise IOError("Too many cosine bins in the tally n. %d of %s" % (tally.tallyNumber, self.mctalFileName))
                        self.line = self.mctalFile.readline()

                if tally.radiograph is True:
                    tally.seg = tally.seg[:-1]
                    tally.seg = tally.seg[:-1] 
                    tally.seg = tally.seg[:-1] 
                    tally.cos = tally.cos[:-1]
                    tally.cos = tally.cos[:-1] 
                    tally.cos = tally.cos[:-1]
                # ERG
                self.line = self.line.split()
                if self.line[0].lower() == "et": tally.ergTC = "t"
                if self.line[0].lower() == "ec": tally.ergTC = "c"
                tally.nErg = int(self.line[1])
                if len(self.line) == 3: tally.ergFlag = int(self.line[2])

                # ENERGY BINS
                self.line = self.mctalFile.readline()
                while self.line[0].lower() != "t":
                        for e in self.line.split():
                                if not tally._insertErg(float(e)):
                                        raise IOError("Too many energy bins in the tally n. %d of %s" % (tally.tallyNumber, self.mctalFileName))
                        self.line = self.mctalFile.readline()

                # TIM
                self.line = self.line.split()
                if self.line[0].lower() == "tt": tally.timTC = "t"
                if self.line[0].lower() == "tc": tally.timTC = "c"
                tally.nTim = int(self.line[1])
                if len(self.line) == 3: tally.timFlag = int(self.line[2])

                # TIME BINS
                self.line = self.mctalFile.readline()
                while self.line.strip().lower() != "vals":
                        for t in self.line.split():
                                if not tally._insertTim(float(t)):
                                        raise IOError("Too many time bins in the tally n. %d of %s" % (tally.tallyNumber, self.mctalFileName))
                        self.line = self.mctalFile.readline()

                # VALS
                f = 1
                Fld = []
                nFld = 0
                tally._initializeValuesVectors()

                nCells = tally._getNbins("f")
                nCora  = tally._getNbins("i")
                nCorb  = tally._getNbins("j")
                nCorc  = tally._getNbins("k")
                nDir   = tally._getNbins("d")
                nUsr   = tally._getNbins("u")
                nSeg   = tally._getNbins("s")
                nMul   = tally._getNbins("m")
                nCos   = tally._getNbins("c")
                nErg   = tally._getNbins("e")
                nTim   = tally._getNbins("t")

                for c in range(nCells):
                        for d in range(nDir):
                                for u in range(nUsr):
                                        for s in range(nSeg):
                                                for m in range(nMul):
                                                        for a in range(nCos): # a is for Angle...forgive me
                                                                for e in range(nErg):
                                                                        for t in range(nTim):
                                                                                for k in range(nCorc):
                                                                                        for j in range(nCorb):
                                                                                                for i in range(nCora):
                                                                                                        if (f > nFld): # f is for Field...again, forgive me
                                                                                                                del Fld
                                                                                                                del self.line
                                                                                                                self.line = self.mctalFile.readline().strip()
                                                                                                                Fld = self.line.split()
                                                                                                                nFld = len(Fld) - 1
                                                                                                                f = 0

                                                                                                        if self.line[0:3] != "tfc":
                                                                                                                # This needs to handle the bug like '8.23798-100'
                                                                                                                try:
                                                                                                                    val = float(Fld[f])
                                                                                                                    err = float(Fld[f+1])
                                                                                                                except ValueError:
                                                                                                                    val = 0
                                                                                                                    err = 0
                                                                                                                    
                                                                                                                if math.isnan(val) or math.isnan(err):
                                                                                                                        self.thereAreNaNs = True
                                                                                                                tally._insertValue(c,d,u,s,m,a,e,t,i,j,k,0,val)
                                                                                                                tally._insertValue(c,d,u,s,m,a,e,t,i,j,k,1,err)

                                                                                                                f += 2

                del Fld

                if tally.mesh == False:
                        # TFC JTF
                        self.line = self.mctalFile.readline().strip().split()
                        if self.line[0] != "tfc":
                                raise IOError("There seem to be more values than expected in tally n. %d of %s" % (tally.tallyNumber, self.mctalFileName))

                        del self.line[0]
                        self.line = [int(i) for i in self.line]

                        if not tally._insertTfcJtf(self.line):
                                raise IOError("Wrong number of TFC jtf elements.")


                        # TFC DAT
                        self.line = self.mctalFile.readline().strip()
                        while "tally" not in self.line and len(self.line) != 0:

                                self.line = self.line.split()

                                tfcDat = []

                                tfcDat.append(int(self.line[0]))
                                try:
                                    val1 = float(self.line[1])
                                except ValueError:
                                    val1 = 0

                                if math.isnan(val1) or math.isnan(float(self.line[2])):
                                        self.thereAreNaNs = True

                                tfcDat.append(val1)
                                tfcDat.append(float(self.line[2]))
                                if len(self.line) == 4:
                                        if math.isnan(val1):
                                                self.thereAreNaNs = True
                                        tfcDat.append(float(self.line[3]))

                                if not tally._insertTfcDat(tfcDat):
                                        raise IOError("Wrong number of elements in TFC data line in the tally n. %d of %s" % (tally.tallyNumber, self.mctalFileName))

                                self.line = self.mctalFile.readline().strip()

                else:
                        while "tally" not in self.line and len(self.line) != 0:
                                self.line = self.mctalFile.readline().strip()


                self.tallies.append(tally)

                if self.line == "":
                        self.line = self.line
                        return True
                elif "tally" in self.line:
                        self.line = self.line.split()
                        return False
   

    def _get_dfs(self, collapse: bool=False
                 ) -> tuple[dict[str, pd.DataFrame], dict[str, pd.DataFrame]]:
        """
        Retrieve and organize mctal data into a DataFrame.

        Parameters
        ----------
        collapse : bool
            collapse the Cell and segments binning in a single Cell-Segment
            one

        Returns
        -------
        tallydata : dict[str, pd.DataFrame]
            organized tally data.
        totalbin : dict[str, pd.DataFrame]
            organized tally data (only total bins).

        """
        tallydata = {}
        totalbin = {}

        for t in self.tallies:
            rows = []

            # --- Reorganize values ---
            # You cannot recover the following from the mctal
            nDir = t._getNbins("d", False)
            nMul = t._getNbins("m", False)
            nSeg = t._getNbins("s", False)  # this can be used

            # Some checks for voids
            binnings = {'cells': t.cells, 'user': t.usr, 'segments': t.seg,
                        'cosine': t.cos, 'energy': t.erg, 'time': t.tim,
                        'cor A': t.cora, 'cor B': t.corb, 'cor C': t.corc}

            # Cells may have a series of zeros, the last one may be for the
            # total
            cells = []
            # last_idx = len(binnings['cells'])-1
            for i, cell in enumerate(binnings['cells']):
                if int(cell) == 0:
                    newval = 'Input '+str(i+1)
                    cells.append(newval)
                # Everything is fine, nothing to modify
                else:
                    # force it to be an int
                    cells.append(int(cell))
            binnings['cells'] = cells

            for name, binning in binnings.items():
                if len(binning) == 0:
                    binnings[name] = [np.nan]
            # Start iteration
            for f, fn in enumerate(binnings['cells']):
                for d in range(nDir):  # Unused
                    for u, un in enumerate(binnings['user']):
                        for sn in range(1, nSeg+1):
                            for m in range(nMul):  # (unused)
                                for c, cn in enumerate(binnings['cosine']):
                                    for e, en in enumerate(binnings['energy']):
                                        for nt, ntn in enumerate(binnings['time']):
                                            for k, kn in enumerate(binnings['cor C']):
                                                for j, jn in enumerate(binnings['cor B']):
                                                    for i, ina in enumerate(binnings['cor A']):
                                                        val = t._getValue(f, d, u, sn-1, m, c, e, nt, i, j, k, 0)
                                                        err = t._getValue(f, d, u, sn-1, m, c, e, nt, i, j, k, 1)
                                                        rows.append([fn, d, un, sn, m, cn, en, ntn, ina, jn, kn, val,err])

                # Only one total bin per cell is admitted
                val = t._getValue(f, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0)
                err = t._getValue(f, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 1)
                if t.timTC is not None:
                    rows.append([fn, d, un, sn, m, cn, en, 'total', ina, jn,
                                 kn, val, err])
                    total = 'Time'

                elif t.ergTC is not None:
                    rows.append([fn, d, un, sn, m, cn, 'total', ntn, ina, jn,
                                 kn, val, err])
                    total = 'Energy'

                elif t.segTC is not None:
                    rows.append([fn, d, un, 'total', m, cn, en, ntn, ina, jn,
                                 kn, val, err])
                    total = 'Segments'

                elif t.cosTC is not None:
                    rows.append([fn, d, un, sn, m, 'total', en, ntn, ina, jn,
                                 kn, val, err])
                    total = 'Cosine'

                elif t.usrTC is not None:
                    rows.append([fn, d, 'total', sn, m, cn, en, ntn, ina, jn,
                                 kn, val, err])
                    total = 'User'

            # --- Build the tally DataFrame ---
            columns = ['Cells', 'Dir', 'User', 'Segments',
                       'Multiplier', 'Cosine', 'Energy', 'Time',
                       'Cor C', 'Cor B', 'Cor A', 'Value', 'Error']
            df = pd.DataFrame(rows, columns=columns)

            # Default drop of multiplier and Dir
            del df['Dir']
            del df['Multiplier']
            # --- Keep only meaningful binning ---
            # Drop NA
            df.dropna(axis=1, inplace=True)
            # Drop constant axes (if len is > 1)
            if len(df) > 1:
                for column in df.columns:
                    if column not in ['Value', 'Error']:
                        firstval = df[column].values[0]
                        # Should work as long as they are the exact same value
                        allequal = (df[column] == firstval).all()
                        if allequal:
                            del df[column]

            # Drop rows if they are exactly the same values
            # (untraced behaviour)
            df.drop_duplicates(inplace=True)

            # The double binning Surfaces/cells with segments can create
            # issues for JADE since if another binning is added
            # (such as energy) it is not supported. Nevertheless,
            # the additional segmentation can be quite useful and this can be
            # collapsed de facto in a single geometrical binning

            if ('Cells' in df.columns and
                'Segments' in df.columns and
                collapse and
                len(df)) > 1:
                # Then we can collapse this in a single geometrical binning
                values = []
                for cell, segment in zip(df.Cells, df.Segments):
                    try:
                        val = str(int(cell))+'-'+str(int(segment))
                    except ValueError:
                         # the cell or segment may not be a number
                         val = '{}-{}'.format(cell, segment)
                    values.append(val)
                df['Cells-Segments'] = values
                # delete the collapsed columns
                del df['Cells']
                del df['Segments']

            # Sub DF containing only total bins
            try:
                dftotal = df[df[total] == 'total']
            except (KeyError, NameError):
                # KeyError : there is no total bin in df
                # NameError: total variable was not defined
                dftotal = None

            tallydata[t.tallyNumber] = df
            totalbin[t.tallyNumber] = dftotal

        return tallydata, totalbin



"""
Parsing of MCNP eeout files.

The main feature is the conversion of the results to a .vtk file.
"""

"""
Copyright 2019 F4E | European Joint Undertaking for ITER and the Development of
Fusion Energy (‘Fusion for Energy’). Licensed under the EUPL, Version 1.2 or - 
as soon they will be approved by the European Commission - subsequent versions
of the EUPL (the “Licence”). You may not use this work except in compliance
with the Licence. You may obtain a copy of the Licence at: 
    https://eupl.eu/1.2/en/
Unless required by applicable law or agreed to in writing, software distributed
under the Licence is distributed on an “AS IS” basis, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the Licence permissions
and limitations under the Licence.
"""

import re
import os
import logging
import pyvista as pv
import numpy as np

from f4enix.constants import PAT_DIGIT, PAT_SPACE, SCIENTIFIC_PAT

# Identifiers
# idNodes = 'NUMBER OF NODES'
# idTets = 'NUMBER OF 1st TETS'
# id2Tets = 'NUMBER OF 2nd TETS'
# idParticlesType = 'PARTICLE LIST'
# idNodeX = 'NODES X'
# idNodeY = 'NODES Y'
# idNodeZ = 'NODES Z'
# idElem = 'ELEMENT TYPE'
# idConn = 'CONNECTIVITY DATA 1ST ORDER TETS ELEMENT ORDERED'
# id2Conn = 'CONNECTIVITY DATA 2ND ORDER TETS ELEMENT ORDERED'
# idNeighbour = 'NEAREST NEIGHBOR DATA 1ST ORDER TETS'
# id2Neighbour = 'NEAREST NEIGHBOR DATA 2ND ORDER TETS'
# Common patterns
PAT_INFO_NAME = re.compile(r'[A-Za-z\s\d]+')
PAT_VALUE = re.compile(r'\s[\s-]\d')
# patNumber = re.compile(r'\d+')
# patNumberSci = re.compile(r'[-+]*\d+.\d+E[+-]\d+')
# patSpace = re.compile(r'\s+')
# patFlux = re.compile(r'd*4$')
# patHeat = re.compile(r'd*6$')
# # special patterns
# patTets = re.compile(r'\d\d+')
TETRA1 = '1st tetra'
TETRA2 = '2nd tetra'


class EEOUT:

    def __init__(self, filepath: os.PathLike) -> None:
        """Representation of an MCNP .eeout output file

        This class is used to parse and manipulate a .eeout file. The main 
        feature is the possibility to convert the results to a .vtk format

        Parameters
        ----------
        filepath : os.PathLike
            path to the .eeout file to be parsed.

        Attributes
        ----------
        filename : str
            name of the original .eeout file
        lines : list[str]
            lines of the file stored in memory
        info : dict
            contains general informations about the file
        n_nodes : int
            number of nodes of the mesh
        n_elem : int
            number of elements of the mesh
        grid : pv.UnstructuredGrid
            vtk object containing all the parsed data from eeout

        Examples
        --------
        Parse an eeout file and export it to .vtu format

        >>> from f4enix.output.eeout import EEOUT
        ... eeout = EEOUT(r'path/to/file.eeout')
        ... eeout.export(r'path/to/output/folder')

        Raises
        ------
        NotImplementedError
            Only 1st and 2nd order tetras are supported
        """
        self.filename = os.path.basename(filepath).split('.')[0]
        logging.info(f'Parsing {self.filename}')
        # store file lines
        lines = []
        with open(filepath, 'r', errors="surrogateescape") as infile:
            for line in infile:
                lines.append(line)
        self.lines = lines

        # Get infos
        self.info = self._read_info()

        # get materials name
        mat_names = self._get_materials_name()

        # Determine mesh elements used. Only first and second order tetras
        # are supported. No mixed formulation are supported
        tetra1 = self.info['NUMBER OF 1st TETS']
        tetra2 = self.info['NUMBER OF 2nd TETS']
        self.n_nodes = self.info['NUMBER OF NODES']
        if tetra1 > 0 and tetra2 == 0:
            self.elem_type = TETRA1
            self.n_elem = tetra1
        elif tetra2 > 0 and tetra1 == 0:
            self.elem_type = TETRA2
            self.n_elem = tetra2
        else:
            raise NotImplementedError(
                'Only 1st and 2nd order tetra are supported. No mixed formulation')

        # get the particle list
        self.p_list = self._read_particle_list()
        assert len(self.p_list) == self.info['NUMBER OF PARTICLES']

        points, idx_elem_type = self._read_nodes_xyz()

        # get the materialss
        material_ids, idx_connect = self._read_material(idx_init=idx_elem_type)

        # TODO the element type should be parsed here in order to allow mixed
        # formulations

        # get connectivity
        elem_connectivity, idx_neigh = self._read_connectivity(
            start_idx=idx_connect)

        # get edits
        edits, idx_centroids = self._read_edits(start_idx=idx_neigh)

        # read density and volumes
        rhos, volumes = self._read_rho_vol(start_index=idx_centroids)

        # build the UM
        if self.elem_type == TETRA1:
            cell_type = np.full(self.n_elem, pv.CellType.TETRA, dtype=np.uint8)
        elif self.elem_type == TETRA2:
            cell_type = np.full(self.n_elem, pv.CellType.TETRA, dtype=np.uint8)
        else:
            raise NotImplementedError(f'{self.elem_type} not implemented')

        grid = pv.UnstructuredGrid(elem_connectivity.ravel(), cell_type,
                                   points)
        grid['material'] = np.vectorize(mat_names.get)(material_ids)
        grid['density'] = rhos
        grid['volume'] = volumes
        for key, field in edits.items():
            grid[f'{key} - value'] = field['values']
            grid[f'{key} - error'] = field['errors']

        self.grid = grid

        logging.info('Parsing completed correctly')

    def export(self, outfolder: os.PathLike, filename: str = None,
               format: str = 'vtu') -> None:
        """Export the eeout to different format.

        at the moment only unstructured mesh vtk (vtu) is the only supported
        format.

        Parameters
        ----------
        outfolder : os.PathLike
            path to the output folder where to export.
        filename : str, optional
            name of the file (without format extension), by default None,
            meaning the the original name of the eeout file will be used.
        format : str, optional
            export format, by default 'vtu' which is only one that is currently
            supported.

        Raises
        ------
        ValueError
            if the output folder does not exists.
        """
        if not os.path.exists(outfolder):
            raise ValueError(f'{outfolder} does not exist')

        if format not in ['vtu']:
            raise NotImplementedError(f'{format} is not a valid format')

        if filename is None:
            filename = self.filename

        outpath = os.path.join(outfolder, filename+'.'+format)
        self.grid.save(outpath)

    def _read_info(self) -> dict:
        read = False
        infos = {}
        for line in self.lines:
            # read infos between n. of particles and number of com edits
            if line.find('NUMBER OF PARTICLES') != -1:
                read = True

            if read:
                info_name = PAT_INFO_NAME.match(line).group().strip()
                value = PAT_DIGIT.findall(line)[-1]
                infos[info_name] = int(value)

            if line.find('NUMBER OF EDITS') != -1:
                return infos

        raise RuntimeError("No 'NUMBER OF EDITS' tag was found")

    def _get_materials_name(self) -> dict[int, str]:
        n_materials = self.info['NUMBER OF MATERIALS']
        for idx, line in enumerate(self.lines):
            pat_flag = re.compile(r'\s+MATERIALS')
            if pat_flag.match(line):
                break

        pat_whatever = re.compile('.+')
        materials = {}
        for i in range(n_materials):
            line = self.lines[i+1+idx]
            mat = pat_whatever.match(line).group().strip()
            materials[i+1] = mat

        return materials

    def _read_particle_list(self) -> int:
        readFlag = False
        for line in self.lines:
            if readFlag:
                particleList = PAT_DIGIT.findall(line)
                particleList = [int(x) for x in particleList]  # get ints
                return particleList

            if line.find('PARTICLE LIST') != -1:
                readFlag = True

        raise RuntimeError("Particles list was not found")

    def _read_material(self, idx_init: int = 0) -> tuple[list[int], int]:
        materialFlag = False
        materialsList = []

        for idx, line in enumerate(self.lines[idx_init:]):
            # No more materials to be read
            if line.find('CONNECTIVITY DATA') != -1:
                del materialsList[-6:]
                return materialsList, idx+idx_init

            if materialFlag:
                strippedline = line.strip()
                matLine = PAT_SPACE.split(strippedline)
                for mat in matLine:
                    materialsList.append(int(mat))

            # trigger
            if line.find('ELEMENT MATERIAL') != -1:
                materialFlag = True

    def _read_nodes_xyz(self) -> tuple[np.ndarray, int]:

        # Reading nodes
        readFlagX = False
        readFlagY = False
        readFlagZ = False

        nodesX = []
        nodesY = []
        nodesZ = []

        pat_trigger_end = re.compile(' ELEMENT TYPE')

        for idx, line in enumerate(self.lines):

            # Reading nodes
            if readFlagX:
                split = PAT_SPACE.split(line)
                for string in split:
                    a = SCIENTIFIC_PAT.search(string)
                    if a is not None:
                        nodesX.append(float(a.group()))

            if readFlagY:
                split = PAT_SPACE.split(line)
                for string in split:
                    a = SCIENTIFIC_PAT.search(string)
                    if a is not None:
                        nodesY.append(float(a.group()))

            if readFlagZ:
                split = PAT_SPACE.split(line)
                for string in split:
                    a = SCIENTIFIC_PAT.search(string)
                    if a is not None:
                        nodesZ.append(float(a.group()))

            if line.find('NODES X') != -1:
                readFlagX = True
            if line.find('NODES Y') != -1:
                readFlagY = True
                readFlagX = False
            if line.find('NODES Z') != -1:
                readFlagZ = True
                readFlagY = False

            # trigger exit
            if pat_trigger_end.match(line) is not None:
                points = np.concatenate([np.array([nodesX]).T,
                                         np.array([nodesY]).T,
                                         np.array([nodesZ]).T], axis=1)
                return points, idx

        raise RuntimeError("Could not find the element tag 'ELEMENT TYPE'")

    def _read_connectivity(self, start_idx: int = 0) -> tuple[np.ndarray, int]:
        data = []
        # Reading nodes
        readFlagCon = False

        for idx, line in enumerate(self.lines[start_idx:]):
            if readFlagCon:
                data.append(line)
            if line.find('CONNECTIVITY DATA') != -1:
                readFlagCon = True
                if line.find('NODE ORDERED') != -1:
                    raise NotImplementedError(
                        'Only element ordered is supported')

            # trigger exit
            if line.find('NEAREST NEIGHBOR DATA') != -1:
                readFlagCon = False
                break

        del data[-1]
        del data[-1]
        # determine number of nodes per element
        if self.elem_type == TETRA1:
            n_nodes = 4
        elif self.elem_type == TETRA2:
            n_nodes = 10

        # first get all values and straigthen the array
        # (solve multiple line issue)
        conn = []
        for string in data:
            newline = string.strip()
            substrings = PAT_SPACE.split(newline)
            for substring in substrings:
                num = int(PAT_DIGIT.search(substring).group())-1
                conn.append(num)

        conn = np.reshape(np.array(conn), (-1, n_nodes))
        # add the n_nodes describer
        describer = np.full((len(conn), 1), n_nodes)

        # finally concat
        connectivity = np.concatenate([describer, conn], axis=1)

        return connectivity, idx+start_idx

    def _read_edits(self, start_idx: int = 0
                    ) -> tuple[dict[str, dict[str, list]], int]:

        pat_particle = re.compile(r'(?<=DATA OUTPUT PARTICLE :)\s+\d+')
        pat_edit = re.compile(r'(?<=EDIT LIST :)\s+\d+')
        pat_type = re.compile(r'(?<=TYPE :)\s+\w+')

        valuesFlag = False

        edits = {}

        for idx, line in enumerate(self.lines[start_idx:]):
            # a new edit is found, initialize it
            if line.find('DATA OUTPUT PARTICLE') != -1:
                par = int(pat_particle.search(line).group())
                edit_num = int(pat_edit.search(line).group())
                edit_type = pat_type.search(line).group().strip()
                current_tally = f'Tally{edit_num}_par{par}_{edit_type}'
                logging.info(f'parsing {current_tally}')
                edits[current_tally] = {}

            # We need to read data (either values or errors)
            if valuesFlag:
                # valid data line
                if PAT_VALUE.match(line) is not None:
                    strippedline = line.strip()
                    values = PAT_SPACE.split(strippedline)
                    for value in values:
                        values_list.append(float(value))
                # no more values in this block
                else:
                    edits[current_tally][value_type] = values_list[1:]
                    valuesFlag = False

            # triggers for following line
            if (line.find('DATA SETS RESULT TIME BIN') != -1 or
                line.find('DATA SETS REL ERROR TIME BIN') != -1):
                valuesFlag = True
                values_list = []
                if 'RESULT' in line:
                    value_type = 'values'
                elif 'REL ERROR' in line:
                    value_type = 'errors'
                else:
                    value_type = None

            # triggers the end of the data section
            if line.find('CENTROIDS X') != -1:
                return edits, idx+start_idx

        raise RuntimeError(
            "Could not parse edits properly, missing 'CENTROIDS X'")

    def _read_rho_vol(self, start_index: int = 0) -> tuple[list[float],
                                                           list[float]]:
        densityFlag = False
        volumeFlag = False
        densityList = []
        volumesList = []
        for line in self.lines[start_index:]:
            if densityFlag or volumeFlag:
                # valid data line
                if PAT_VALUE.match(line) is not None:
                    strippedline = line.strip()
                    splitline = PAT_SPACE.split(strippedline)
                    for value in splitline:
                        if densityFlag:
                            densityList.append(float(value))
                        elif volumeFlag:
                            volumesList.append(float(value))

            if line.find('DENSITY') != -1:
                densityFlag = True
            if line.find('VOLUMES') != -1:
                volumeFlag = True
                densityFlag = False

        return densityList, volumesList

    def __repr__(self) -> str:
        return str(self.info)

    def __str__(self) -> str:
        return str(self.info)



"""
Auxiliary functions used by different modules

"""
"""
Copyright 2019 F4E | European Joint Undertaking for ITER and the Development of
Fusion Energy (‘Fusion for Energy’). Licensed under the EUPL, Version 1.2 or - 
as soon they will be approved by the European Commission - subsequent versions
of the EUPL (the “Licence”). You may not use this work except in compliance
with the Licence. You may obtain a copy of the Licence at: 
    https://eupl.eu/1.2/en/ 
Unless required by applicable law or agreed to in writing, software distributed
under the Licence is distributed on an “AS IS” basis, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the Licence permissions
and limitations under the Licence.
"""

from numjuggler.parser import Card
from f4enix.constants import PAT_COMMENT, PAT_DOLLAR_COMMENT, PAT_COMMENT_TEXT
import re
import os

_surrogates = re.compile(r"[\uDC80-\uDCFF]")


def get_comments(card: Card) -> str:
    """Analyze the lines of a card and get the comments

    Parameters
    ----------
    card : Card
        card to be analyzed

    Returns
    -------
    str
        comments as a single string
    """
    comments = ''
    for line in card.lines:
        # check if either dollar or c comments are matched and store them
        c_comm = PAT_COMMENT_TEXT.match(line)
        d_comm = PAT_DOLLAR_COMMENT.search(line)

        for match in [c_comm, d_comm]:
            if match is not None:
                comments = comments+match.group()

    return comments


def _detect_decoding_errors_line(l, _s=_surrogates.finditer):  # pragma: no cover
    """Return decoding errors in a line of text

    Works with text lines decoded with the surrogateescape
    error handler.

    Returns a list of (pos, byte) tuples

    """
    # DC80 - DCFF encode bad bytes 80-FF
    return [(m.start(), bytes([ord(m.group()) - 0xDC00]))
            for m in _s(l)]


def debug_file_unicode(file: os.PathLike) -> str:  # pragma: no cover
    """given a file prints the unicode error that were found and at what
    line

    Parameters
    ----------
    file : os.PathLike
        file to be debugged

    Returns
    -------
    str
        bug found
    """
    txt = ''
    with open(file, 'r', errors="surrogateescape") as f:
        for i, line in enumerate(f, 1):
            errors = _detect_decoding_errors_line(line)
            if errors:
                txt += (f"Found errors on line {i}:")
                for (col, b) in errors:
                    txt += (f" {col + 1:2d}: {b[0]:02x}")
                txt += '\n'
    return txt



"""This module is related to the parsing of D1S-UNED meshinfo files.

"""
from __future__ import annotations
"""
Copyright 2019 F4E | European Joint Undertaking for ITER and the Development of
Fusion Energy (‘Fusion for Energy’). Licensed under the EUPL, Version 1.2 or -
as soon they will be approved by the European Commission - subsequent versions
of the EUPL (the “Licence”). You may not use this work except in compliance
with the Licence. You may obtain a copy of the Licence at:
    https://eupl.eu/1.2/en/  
Unless required by applicable law or agreed to in writing, software distributed
under the Licence is distributed on an “AS IS” basis, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the Licence permissions
and limitations under the Licence.
"""

import numpy as np
import os
from typing import BinaryIO, Dict
import matplotlib.pyplot as plt
import matplotlib.colors as colors

BYTE = np.byte
CHAR = np.char
INT = np.int32
FLOAT = np.float64
LONG = np.int64


class RSSA:
    def __init__(self, file: os.PathLike) -> None:
        """
        Representation of a RSSA file.

        Parameters
        ----------
        file : os.PathLike
            path to the rssa file to be parsed

        Attributes
        ----------
        self.parameters: dict
            {np1,  # Number of histories of the simulation, given as a negative number
            nrss,  # Number of tracks recorded
            nrcd,  # Number of values recorded for each particle, it should be 11
            njsw,  # Number of surfaces in JASW
            niss,  # Number of different histories that reached the SSW surfaces
            niwr,  # Number of cells in RSSA file
            mipts,  # Source particle type
            kjaq,  # Flag for macrobodies surfaces
            surfaces}
        self.tracks: np.ndarray
            each index of the array has 11 values
            0 a,  # History number of the particle, negative if uncollided
            1 b,  # Packed variable, the sign is the sign of the third direction cosine, starts with 8 = neutron, 16 = photon
            2 wgt,
            3 erg,
            4 tme,
            5 x,
            6 y,
            7 z,
            8 u,  # Particle direction cosine with X-axis
            9 v,  # Particle direction cosine with Y-axis, to calculate w (Z-axis) use the sign from b
            10 c  # Surface id
        filename : str
            name of the original file

        Examples
        --------
        Load a RSSA file and get some info

        >>> from f4enix.output.rssa import RSSA
        ... my_rssa = RSSA('small_cyl.w')
        ... print(my_rssa)
        RSSA file small_cyl was recorded using the following surfaces:
        Surface id: 1
        The surface type is a cylinder with a radius of 30.00
        The total amount of tracks recorded is 72083, of which 72083 were neutrons and 0 were photons.
        The simulation that produced this RSSA run 100000 histories
        The amount of independent histories that reached the RSSA surfaces was 70797.

        """
        self.filename = os.path.basename(file).split('.')[0]
        self.parameters, self.tracks = self._read_rssa(file)

    def __repr__(self) -> str:
        return self._get_info()

    def __str__(self) -> str:
        return self._get_info()

    @staticmethod
    def _read_rssa(filename: str) -> tuple[dict, np.array]:
        with open(filename, 'rb') as infile:
            # This parameters hold information like the amount of histories or
            # the amount of tracks recorded
            parameters = _read_header(infile)
            tracks = _read_tracks(infile)

        return parameters, tracks

    def _calculate_grid_axes_cyl(self, z_int: int, theta_int: int,
                                 mask: np.ndarray = None):
        """The axes are calculated without taking into account the type of
        particle"""
        if mask is None:
            mask = np.where(self.x)[0]

        z_min = self.z[mask].min()
        z_max = self.z[mask].max()

        thetas = np.angle(self.x[mask] + 1j * self.y[mask])  # in radians
        theta_min = np.min(thetas)
        theta_max = np.max(thetas)

        z_axis = np.linspace(z_min, z_max, z_int + 1)
        theta_axis = np.linspace(theta_min, theta_max, theta_int + 1)
        return z_axis, theta_axis

    def _generate_figures_current_cyl(
            self, particle: str = 'n', z_int: int = 10, theta_int: int = 10,
            source_intensity: float = 1.7757e20,
            value_range: tuple[float, float] = None,
            mask: np.ndarray = None) -> tuple[plt.Figure, plt.Figure]:

        particle_mask = self._get_particle_mask(particle)
        if mask is not None:  # Apply to the mask a geom filter done earlier
            particle_mask = np.intersect1d(particle_mask, mask)

        # Create the empty grid
        z_axis, theta_axis = self._calculate_grid_axes_cyl(z_int, theta_int,
                                                           particle_mask)
        grid_values = np.zeros((z_int, theta_int))
        error_grid = np.zeros((z_int, theta_int))

        # Calculate the indices at both axis of the grids for each track
        z_idx = _closest(z_axis, self.z[particle_mask])
        # in radians
        thetas = np.angle(self.x[particle_mask] + 1j * self.y[particle_mask])
        theta_idx = _closest(theta_axis, thetas)
        del thetas  # To relax a bit memory constraints

        # Populate both grids
        np.add.at(grid_values, (z_idx, theta_idx), self.wgt[particle_mask])
        np.add.at(error_grid, (z_idx, theta_idx), 1)

        # Normalize values
        # radius of the cylinder
        radius = np.linalg.norm([self.x[0], self.y[0]])
        extent = [radius*theta_axis[0], radius*theta_axis[-1],
                  z_axis[0], z_axis[-1]]  # used in the plotting
        area = abs(radius*(theta_axis[1]-theta_axis[0])*(z_axis[1]-z_axis[0]))
        grid_values /= area
        grid_values /= abs(self.parameters['np1'])  # grid /= nps
        grid_values *= source_intensity
        # Give a relative error of 1 to empty voxels
        error_grid[np.where(error_grid == 0)] = 1
        error_grid = 1 / np.sqrt(error_grid)

        # Generate the figure values
        figure_values: plt.Figure = plt.figure()
        ax_values: plt.Axes = figure_values.add_subplot()
        ax_values.set_xlabel("Perimeter of the cylinder (cm)")
        ax_values.set_ylabel("Z (cm)")
        if particle == 'n':
            ax_values.set_title("Neutron current through the surface [#/cm2/s]")
        else:
            ax_values.set_title("Photon current through the surface [#/cm2/s]")
        ax_values.imshow(grid_values, origin='lower', extent=extent)
        # Set the colors to log range
        if value_range is not None:
            log_max = int(np.log10(value_range[1]))
            log_min = int(np.log10(value_range[0]))
        else:
            # The +1 is needed so max=1234 => 10.000
            log_max = int(np.log10(grid_values.max())) + 1
            log_min = log_max - 10
        decades = log_max - log_min
        image_values = ax_values.images[0]
        # set the colormap and number of bins
        image_values.cmap = plt.get_cmap('jet', decades)
        # set the scale as log
        image_values.norm = colors.LogNorm(np.power(10., log_min),
                                           np.power(10., log_max))
        _color_bar = figure_values.colorbar(image_values,
                                            orientation='horizontal')

        # Generate the errors figure
        figure_errors: plt.Figure = plt.figure()
        ax_errors: plt.Axes = figure_errors.add_subplot()
        ax_errors.set_xlabel("Perimeter of the cylinder (cm)")
        ax_errors.set_ylabel("Z (cm)")
        ax_errors.set_title("Relative error as 1/sqrt(N)")
        norm = colors.Normalize(0, 1)
        color_map = plt.get_cmap('jet', 10)
        ax_errors.imshow(error_grid, cmap=color_map, norm=norm,
                         origin='lower', extent=extent)
        image_errors = ax_errors.images[0]
        figure_errors.colorbar(image_errors, orientation='horizontal')

        # Print information about the grid
        print(f"The area of a cell is {area:.2f}cm2")
        print(f"The resolution is {radius * (theta_axis[1] - theta_axis[0]):.2f}cm x {z_axis[1] - z_axis[0]:.2f}cm")

        return figure_values, figure_errors

    def plot_cyl(self,
                 particle: str = 'n',
                 z_int: int = 10,
                 theta_int: int = 10,
                 norm: float = 1,
                 value_range: tuple[float, float] = None,
                 x_range: tuple[float, float] = None,
                 z_range: tuple[float, float] = None,
                 outfolder: os.PathLike = None,
                 filename: str = None) -> tuple[plt.Figure, plt.Figure]:
        """Plot the cylinder surface of RSSA.

        Parameters
        ----------
        particle : str, optional
            either neutron 'n' or photon 'p', by default 'n'
        z_int : int, optional
            z grid discretizations, by default 10
        theta_int : int, optional
            theta grid discretizations, by default 10
        norm : float, optional
            normalization to be applied to the results, by default 1
        value_range : tuple[float, float], optional
            min and max values to plot, by default None
        x_range : tuple[float, float], optional
            min x and min z to plot, by default None
        z_range : tuple[float, float], optional
            min z and max z to plot, by default None
        outfolder : os.PathLike, optional
            path to the output folder, default is None causing the plot not
            to be saved
        filename : str, optional
            ovverrides default name for the plot, by default None

        Returns
        -------
        tuple[plt.Figure, plt.Figure]
            Values and Error figures
        """
        # Filter tracks location if necessary
        mask = None
        if z_range is not None:
            mask1 = np.where(self.z > z_range[0])
            mask2 = np.where(self.z < z_range[1])
            mask = np.intersect1d(mask1, mask2)
            del mask1, mask2
        if x_range is not None:
            thetas = np.angle(self.x + 1j * self.y)  # in radians
            # radius of the cylinder
            radius = np.linalg.norm([self.x[0], self.y[0]])
            # Perimeter of the cylinder values, x values in the plot x-axis
            x_values = radius*thetas
            mask1 = np.where(x_values > x_range[0])
            mask2 = np.where(x_values < x_range[1])
            mask = np.intersect1d(mask, mask1)
            mask = np.intersect1d(mask, mask2)
            del mask1, mask2, thetas, radius, x_values

        figure_values, figure_errors = self._generate_figures_current_cyl(
            particle=particle,
            z_int=z_int,
            theta_int=theta_int,
            source_intensity=norm,
            value_range=value_range,
            mask=mask)

        if outfolder is not None:
            if filename is None:
                filename = '{}_{}_z{}_theta{}'.format(self.filename, particle,
                                                      z_int, theta_int)

            outval = os.path.join(outfolder, filename+'.jpeg')
            outerr = os.path.join(outfolder, filename+'_errors.jpeg')

            figure_values.savefig(outval, format='jpeg', dpi=1200)
            figure_errors.savefig(outerr + '_errors.jpeg', format='jpeg',
                                  dpi=1200)
        return figure_values, figure_errors

    def _get_info(self) -> str:
        info = f'RSSA file {self.filename} was recorded using the following surfaces:\n'
        for surface in self.parameters['surfaces']:
            info += f'  Surface id: {surface["id"]}\n'

        sur_type = self.type
        if sur_type == 'cyl':
            info += f'The surface type is a cylinder with a radius of {np.linalg.norm([self.x[0], self.y[0]]):.2f}\n'
        elif sur_type == 'plane':
            info += f'The surface type is a plane...\n'

        n_tracks = self.tracks[self.mask_neutron_tracks].shape[0]
        p_tracks = self.tracks[self.mask_photon_tracks].shape[0]
        info += f'The total amount of tracks recorded is {self.parameters["nrss"]}, of which {n_tracks} were neutrons' \
                f' and {p_tracks} were photons.\n'

        info += f'The simulation that produced this RSSA run {np.abs(self.parameters["np1"])} histories\n' \
                f'The amount of independent histories that reached the RSSA surfaces was {self.parameters["niss"]}.\n'
        return info

    @property
    def mask_neutron_tracks(self) -> np.ndarray:
        '''
        The neutron tracks
        '''
        # Get all the bitarrays and don't pay attention to the sign
        bitarrays = np.abs(self.tracks[:, 1])
        # Neutrons start with 8 and photons with 16 followed by 1e8
        n_tracks = np.where(bitarrays < 9e8)[0]
        return n_tracks

    @property
    def mask_photon_tracks(self) -> np.array:
        '''
        The photon tracks
        '''
        # Get all the bitarrays and don't pay attention to the sign
        bitarrays = np.abs(self.tracks[:, 1])
        # Neutrons start with 8 and photons with 16 followed by 1e8
        p_tracks = np.where(bitarrays >= 9e8)[0]
        return p_tracks

    def _get_particle_mask(self, particle: str):
        if particle == 'n':
            return self.mask_neutron_tracks
        elif particle == 'p':
            return self.mask_photon_tracks
        else:
            raise ValueError(f"Particle was {particle}, not n or p...")

    @property
    def x(self) -> np.ndarray:
        '''
        x positions
        '''
        return self.tracks[:, 5]

    @property
    def y(self) -> np.ndarray:
        '''
        y positions
        '''
        return self.tracks[:, 6]

    @property
    def z(self) -> np.ndarray:
        '''
        z positions
        '''
        return self.tracks[:, 7]

    @property
    def wgt(self) -> np.ndarray:
        '''
        weights
        '''
        return self.tracks[:, 2]

    @property
    def energies(self) -> np.ndarray:
        '''
        energies
        '''
        return self.tracks[:, 3]

    @property
    def histories(self) -> np.ndarray:
        '''
        histories
        '''
        return np.abs(self.tracks[:, 0])

    @property
    def type(self):
        '''
        RSSA surface type
        '''
        # If there are more than 1 surf we cannot say if it is a cyl or a plane
        if len(self.parameters['surfaces']) > 1:
            return 'multiple'
        # Assume it is a cyl and calculate the radius of its tracks
        # if they are different it is a plane
        radius = np.sqrt(self.x ** 2 + self.y ** 2)
        if np.std(radius) < 1e-4:
            return 'cyl'
        else:
            return 'plane'


# --- Helper Functions ---
def _read_fortran_record(file: BinaryIO):
    count_1 = np.fromfile(file, INT, 1)[0]
    data = np.fromfile(file, np.byte, count_1)
    count_2 = np.fromfile(file, INT, 1)[0]
    if count_1 != count_2:
        raise ValueError(
            "The integers that go before and after the Fortran record are not equal...")
    return data


def _read_header(file: BinaryIO) -> Dict:
    # First record
    data = _read_fortran_record(file)
    # The first line of the file with information like the code version, date and title
    format_record_id = data.tobytes().decode('UTF-8')
    if 'd1suned' in format_record_id:
        # TODO: we could parse and store information like datetime and title
        _last_dump = np.frombuffer(data[-4:], INT)
        pass
    else:
        raise NotImplementedError(f'The code that generated this RSSA file has not been implemented'
                                  f' in this parser, see the code here: {format_record_id}...')

    # Second record
    data = _read_fortran_record(file)
    np1 = np.frombuffer(data, LONG, 1, 0)[0]
    nrss = np.frombuffer(data, LONG, 1, 8)[0]
    nrcd = np.frombuffer(data, INT, 1, 16)[0]
    njsw = np.frombuffer(data, INT, 1, 20)[0]
    niss = np.frombuffer(data, LONG, 1, 24)[0]
    if nrcd != 11:
        raise NotImplementedError(
            f"The amount of values recorded for each particle should be 11 instead of {nrcd}...")

    # Third record
    if np1 < 0:
        data = _read_fortran_record(file)
        niwr, mipts, kjaq = np.frombuffer(data, INT, 3)
    else:
        raise NotImplementedError(
            "The np1 value is not negative, as far as we understand it should be negative...")

    # Fourth record
    surfaces = []
    for i in range(njsw):
        data = _read_fortran_record(file)
        surface = {'id': np.frombuffer(data, INT, 1, 0)[0]}
        if kjaq == 1:
            surface['info'] = np.frombuffer(data, INT, 1, 4)[0]
        else:
            surface['info'] = -1
        surface['type'] = np.frombuffer(data, INT, 1, 8)[0]
        surface['num_params'] = np.frombuffer(data, INT, 1, 12)[0]
        surface['params'] = np.frombuffer(data, INT, offset=16)
        surfaces.append(surface)

    # we read any extra records as determined by njsw+niwr...
    # no known case of their actual utility is known currently
    for j in range(njsw, njsw + niwr):
        _data = _read_fortran_record(file)
        raise NotImplementedError(
            'njsw + niwr values are bigger than njsw, behavior not explained')

    # Summary record
    _data = _read_fortran_record(file)
    # Summary record not processed, its information does not interest us for now

    parameters = {'np1': np1,  # Number of histories of the simulation, given as a negative number
                  'nrss': nrss,  # Number of tracks recorded
                  'nrcd': nrcd,  # Number of values recorded for each particle, it should be 11
                  'njsw': njsw,  # Number of surfaces in JASW
                  'niss': niss,  # Number of different histories that reached the SSW surfaces
                  'niwr': niwr,  # Number of cells in RSSA file
                  'mipts': mipts,  # Source particle type
                  'kjaq': kjaq,  # Flag for macrobodies surfaces
                  'surfaces': surfaces}
    return parameters


def _read_tracks(file: BinaryIO) -> np.ndarray:
    # Particle records
    # Read the whole remaining of the file at once, store all the bytes as a 1D numpy array
    data = np.fromfile(file, BYTE)
    # Reshape the array so each index holds the information of a single particle, we can do this because we
    #  know that the particle records have always the same length, 96 bytes
    data = data.reshape(-1, 96)
    # Remove the first and last 4 bytes, these are two integers that tell the record is 88 bytes long
    data = data[:, 4:-4]
    # Convert the array into a 1D array of float numbers instead of simply bytes
    data = np.frombuffer(data.flatten(), FLOAT)
    # Reshape the array so each index holds the information of a single particle, all the data is already converted
    #  from bytes to floats
    data = data.reshape(-1, 11)
    return data


def _closest(axis, values) -> np.array:
    """
    Given an axis (list of ordered equidistant values) and a value list,
    returns the indexes of the axis that are
     closest to the values.
    E.g. axis = [1, 2, 3, 4] value=[2.23, 4.98] result=[1, 2]
    """
    l0 = axis[0]
    lm = axis[1] - axis[0]
    idx = np.array((values - l0) / float(lm))
    idx = idx.astype(int)
    # Limit case where the value lies at the end of the axis
    idx[idx == len(axis) - 1] = len(axis) - 2

    return idx



"""
Parsing of D1S-UNED additional files

Parsers for the irradiation and reaction files necessary for Direct-1-Step
calculation using D1S-UNED.
"""
from __future__ import annotations
"""
Copyright 2019 F4E | European Joint Undertaking for ITER and the Development of
Fusion Energy (‘Fusion for Energy’). Licensed under the EUPL, Version 1.2 or - 
as soon they will be approved by the European Commission - subsequent versions
of the EUPL (the “Licence”). You may not use this work except in compliance
with the Licence. You may obtain a copy of the Licence at:
    https://eupl.eu/1.2/en/
Unless required by applicable law or agreed to in writing, software distributed
under the Licence is distributed on an “AS IS” basis, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the Licence permissions
and limitations under the Licence.
"""

import re
import os
import logging

from f4enix.constants import PAT_BLANK, PAT_COMMENT, PAT_SPACE
from f4enix.input.libmanager import LibManager

# PAT_COMMENT = re.compile('[Cc]+')

REACFORMAT = '{:>13s}{:>7s}{:>9s}{:>40s}'


class IrradiationFile:

    def __init__(self, nsc: int, irr_schedules: list[Irradiation],
                 header: str = None,
                 formatting: list[int] = [8, 14, 13, 9],
                 name: str = 'irrad') -> None:
        """
        Object representing an irradiation D1S-UNED file.

        It is built as a container of single irradiation object

        Parameters
        ----------
        nsc : int
            number of irradiation schedule.
        irr_schedules : list of Irradiation object
            contains all irradiation objects.
        header : str, optional
            Header of the file. The default is None.
        formatting : list of int, optional
            fwf values for the output columns. The default is [8, 14, 13, 9].
        name : str, optional
            name of the file. The default is 'irrad'.

        Attributes
        ----------
        nsc : int
            number of irradiation schedule.
        irr_schedules : list of Irradiation object
            contains all irradiation objects.
        header : str, optional
            Header of the file. The default is None.
        formatting : list of int, optional
            fwf values for the output columns. The default is [8, 14, 13, 9].
        name : str, optional
            name of the file. The default is 'irrad'.

        Examples
        --------
        Some usage examples

        >>> # parse an existing file
        ... irrad_file = IrradiationFile.from_text('irr_test')
        ... # get the list of irradiation schedules
        ... irrad_file.irr_schedules
        [['24051', '2.896e-07', '5.982e+00', '5.697e+00', 'Cr51'],
         ['25054', '2.570e-08', '5.881e+00', '1.829e+00', 'Mn54'],
         ['26055', '8.031e-09', '4.487e+00', '6.364e-01', 'Fe55']]

        >>> # auxiliary method to retrieve a specific irradiation
        ... print(irrad_file.get_irrad('24051'))
        Daughter: 24051
        lambda [1/s]: 2.896e-07
        times: ['5.982e+00', '5.697e+00']
        comment: Cr51

        >>> # auxiliary method to get all daughters
        ... print(irrad_file.get_daughters())
        ['24051', '25054', '26055']

        Returns
        -------
        None.

        """
        self.nsc = nsc
        self.irr_schedules = irr_schedules
        self.header = header
        self.formatting = formatting

        # Compute irradiation header
        w1 = str(formatting[0])
        w2 = str(formatting[1])
        w3 = str(formatting[2])
        w4 = str(formatting[3])

        head = '{:>'+w1+'s}{:>'+w2+'s}{:>'
        for i in range(nsc):
            head += w3+'s}{:>'

        head += w4+'s}'

        self._irrformat = head
        self.name = name

    def get_daughters(self) -> list[str]:
        """
        Get a list of all daughters among all irradiation files

        Returns
        -------
        list[str]
            list of daughters.

        """
        # Get the list of daughters
        daughters = []
        for irradiation in self.irr_schedules:
            daughters.append(irradiation.daughter)

        return daughters

    def get_irrad(self, daughter: str) -> Irradiation | None:
        """
        Return the irradiation correspondent to the daughter

        Parameters
        ----------
        daughter : str
            (e.g. '24051').

        Returns
        -------
        Irradiation | None
            Returns the irradiation corresponding to the daughter.
            If no irradiation is found returns None.

        """
        for irradiation in self.irr_schedules:
            if daughter == irradiation.daughter:
                return irradiation

        return None

    @classmethod
    def from_text(cls, filepath: os.PathLike) -> IrradiationFile:
        """
        Initialize an IrradiationFile object directly parsing and existing
        irradiation file

        Parameters
        ----------
        cls : TYPE
            DESCRIPTION.
        filepath : os.PathLike
            path to the existing irradiation file.

        Returns
        -------
        None.

        """
        logging.info('Parsing {}'.format(filepath))
        pat_nsc = re.compile('(?i)(nsc)')
        pat_num = re.compile(r'\d+')
        # name = os.path.basename(filepath)
        with open(filepath, 'r') as infile:
            inheader = True
            header = ''
            irr_schedules = []
            for line in infile:
                # check if we need to exit header mode
                # it my also happen that there is no header
                if pat_nsc.match(line) is not None:
                    nsc = int(pat_num.search(line).group())
                    inheader = False
                # If in header keep reading header
                elif inheader:
                    header += line
                # data
                else:
                    # Avoid comments and blank lines
                    if (PAT_BLANK.match(line) is None and
                            PAT_COMMENT.match(line) is None):

                        irr_schedules.append(Irradiation.from_text(line, nsc))

        logging.info('{} correctly parsed'.format(filepath))

        return cls(nsc, irr_schedules, header=header)

    def write(self, path: os.PathLike) -> None:
        """
        Write the D1S irradiation file

        Parameters
        ----------
        path : os.PathLike
            output path where to save the file (only directory). self.name will
            be used as output file name

        Returns
        -------
        None.

        """
        filepath = os.path.join(path, self.name)
        with open(filepath, 'w') as outfile:
            if self.header is not None:
                outfile.write(self.header)
            # write nsc
            outfile.write('nsc '+str(self.nsc)+'\n')

            # --- Write irradiation schedules ---
            # write header
            args = ['Daught.', 'lambda(1/s)']
            for i in range(self.nsc):
                args.append('time_fact_'+str(i+1))
            args.append('comments')
            outfile.write('C '+self._irrformat.format(*args)+'\n')

            # write schedules
            for schedule in self.irr_schedules:
                args = schedule._get_format_args()
                outfile.write(self._irrformat.format(*args)+'\n')

        logging.info('Irradiation file written at {}'.format(outfile))


class Irradiation:
    def __init__(self, daughter: str, lambd: str, times: list[str],
                 comment: str = None) -> None:
        """
        Irradiation object

        Parameters
        ----------
        daughter : str
            daughter nuclide (e.g. 24051).
        lambd : str
            disintegration constant [1/s].
        times : list of strings
            time correction factors.
        comment : str, optional
            comment to the irradiation. The default is None.

        Attributes
        ----------
        daughter : str
            daughter nuclide (e.g. 24051).
        lambd : str
            disintegration constant [1/s].
        times : list of strings
            time correction factors.
        comment : str, optional
            comment to the irradiation. The default is None.

        Returns
        -------
        None.

        """
        self.daughter = daughter
        self.lambd = lambd
        self.times = times
        self.comment = comment

    def __eq__(self, other) -> bool:
        """
        Get a more appropriate equivalence function. Two irradiation are equal
        if they have the same daughter, lambda and correction factors

        """
        if isinstance(other, Irradiation):
            daugther_eq = self.daughter == other.daughter
            lamb_eq = float(self.lambd) == float(other.lambd)
            if len(self.times) == len(other.times):
                times_eq = True
                for time1, time2 in zip(self.times, other.times):
                    if float(time1) != float(time2):
                        times_eq = False
            else:
                times_eq = False

            condition = (daugther_eq and lamb_eq and times_eq)

            return condition
        else:
            return False

    @classmethod
    def from_text(cls, text: str, nsc: int) -> Irradiation:
        """
        Parse a single irradiation

        Parameters
        ----------
        cls : TYPE
            DESCRIPTION.
        text : str
            text to be parsed.
        nsc : int
            number of irradiation schedule.

        Returns
        -------
        Irradiation
            Instance of irradiation object.

        """
        pieces = PAT_SPACE.split(text)
        # Check for empty start
        if pieces[0] == '':
            pieces.pop(0)

        daughter = pieces[0]
        lambd = pieces[1]
        times = []
        # Get all decay times
        j = 2
        for i in range(nsc):
            times.append(pieces[j])
            j += 1
        # Get comment
        comment = ''
        try:
            for piece in pieces[j:]:
                comment += ' '+piece
        except IndexError:
            comment = None

        if comment == '':
            comment = None
        else:
            comment = comment.strip()

        return cls(daughter, lambd, times, comment=comment)

    def _get_format_args(self) -> list:
        args = [self.daughter, self.lambd]
        for time in self.times:
            args.append(time)
        args.append(self.comment)
        return args

    def _print(self) -> str:
        text = """
Daughter: {}
lambda [1/s]: {}
times: {}
comment: {}
""".format(self.daughter, self.lambd, self.times, self.comment)

        return text

    def __repr__(self) -> str:
        return str(self._get_format_args())

    def __str__(self) -> str:
        return self._print()


class ReactionFile:
    def __init__(self, reactions: list[Reaction], name: str = 'react') -> None:
        """
        Reaction file object

        Parameters
        ----------
        reactions : list[Reaction]
            contains all reaction objects contained in the file.
        name : name, optional
            file name. The default is 'react'.

        Examples
        --------
        It is possible to change the libraries of a reaction file

        >>> from f4enix.input.d1suned import ReactionFile
        ... reac_file = ReactionFile.from_text('reac_fe')
        ... reac_file.change_lib('98c')

        and obtain a list of the parents

        >>> reac_file.get_parents()
        ['26054', '26056', '26057', '26058']

        Returns
        -------
        None.

        """
        self.reactions = reactions
        self.name = name

    @classmethod
    def from_text(cls, filepath: os.PathLike) -> ReactionFile:
        """
        Generate a reaction file directly from text file

        Parameters
        ----------
        cls : TYPE
            DESCRIPTION.
        filepath : os.PathLike
            file to read.

        Returns
        -------
        ReactionFile
            Reaction File Object.

        """
        # read all reactions
        logging.info('Parsing {}'.format(filepath))
        reactions = []
        with open(filepath, 'r') as infile:
            for line in infile:
                # Ignore if it is a blank line or a full line comment
                if (PAT_BLANK.match(line) is None and
                        PAT_COMMENT.match(line) is None):
                    # parse reactions
                    reaction = Reaction.from_text(line)
                    reactions.append(reaction)
        logging.info('{} correctly parsed'.format(filepath))

        return cls(reactions)  # , name=os.path.basename(filepath))

    def get_parents(self) -> set[str]:
        """
        Get a list of all parents

        Returns
        -------
        set[str]
            list of parents from all reactions

        """
        parents = []
        for reaction in self.reactions:
            parent = reaction.parent.split('.')[0]
            if parent not in parents:
                parents.append(parent)
        return sorted(set(parents))

    def change_lib(self, newlib: str, libmanager: LibManager = None):
        """
        change the parent library tag of the reactions. If no libmanager is
        provided, the check on the availability of the parent in the xsdir
        file will be not performed.

        Parameters
        ----------
        newlib : str
            (e.g. 31c).
        libmanager : LibManager, optional
            Object managing library operations. The default is None.

        Returns
        -------
        None.

        """
        # Correctly parse the lib input. It may be a dic than only the
        # first dic value needs to be cosidered
        pat_libs = re.compile(r'"\d\d[a-zA-Z]"')
        if newlib[0] == '{':
            libs = pat_libs.findall(newlib)
            lib = libs[1][1:-1]
        else:
            lib = newlib

        # actual translation
        for reaction in self.reactions:
            # Insert here a check that the parent isotope is available
            if libmanager is None:
                reaction.change_lib(lib)
            else:
                # get the available libraries for the parent
                zaid = reaction.parent.split('.')[0]
                libs = libmanager.check4zaid(zaid)
                if newlib in libs:
                    reaction.change_lib(lib)
                else:
                    warning = '{} is not available in xsdir, not translated'
                    logging.warning(warning.format(zaid))

    def write(self, path: os.PathLike) -> None:
        """
        write formatted reaction file

        Parameters
        ----------
        path : os.PathLike
            path to the output file (only dir).

        Returns
        -------
        None.

        """
        filepath = os.path.join(path, self.name)
        with open(filepath, 'w') as outfile:
            for reaction in self.reactions:
                outfile.write(REACFORMAT.format(*reaction._get_text())+'\n')
        logging.info('Reaction file written at {}'.format(outfile))

    def _print(self) -> str:
        text = REACFORMAT.format('Parent', 'MT', 'Daughter', 'Comment')+'\n'
        for reaction in self.reactions:
            text = text+REACFORMAT.format(*reaction._get_text())+'\n'
        return text

    def __repr__(self) -> str:
        return self._print()

    def __str__(self) -> str:
        return self._print()


class Reaction:
    def __init__(self, parent: str, MT: int | str, daughter: str,
                 comment: str = None) -> None:
        """
        Represents a single reaction of the reaction file

        Parameters
        ----------
        parent : str
            parent nuclide ZZAAA.XXc representing stable isotope to be
            activated. ZZ and AAA represent the atomic and mass number and
            extension XX, is the extension number of the modified D1S library.
        MT : int | str
            integer, reaction type (ENDF definition, e.g. 102).
        daughter : str
            integer, tag of the daughter nuclide. The value could be
            defined as ZZAAA of daughter nuclide, but any other identification
            type (with integer value) can be used.
        comment : str, optional
            comment to the reaction. The default is None.

        Attributes
        ----------
        parent : str
            parent nuclide ZZAAA.XXc representing stable isotope to be
            activated. ZZ and AAA represent the atomic and mass number and
            extension XX, is the extension number of the modified D1S library.
        MT : str
            integer, reaction type (ENDF definition, e.g. '102').
        daughter : str
            integer, tag of the daughter nuclide. The value could be
            defined as ZZAAA of daughter nuclide, but any other identification
            type (with integer value) can be used.
        comment : str, optional
            comment to the reaction. The default is None.

        Returns
        -------
        None.

        """
        self.parent = parent
        self.MT = str(int(MT))
        self.daughter = daughter
        self.comment = comment

    def change_lib(self, newlib: str) -> None:
        """
        Change the library tag

        Parameters
        ----------
        newlib : str
            library extension as used in xsdir format (e.g. 00c).

        Returns
        -------
        None.

        """
        pieces = self.parent.split('.')
        # Override lib
        self.parent = pieces[0]+'.'+newlib

    def _get_text(self) -> list[str]:
        """
        Generate the reaction text

        Returns
        -------
        text : str
            reaction text for D1S input.

        """
        # compute text
        textpieces = [self.parent, self.MT, self.daughter]
        if self.comment is None:
            comment = ''
        else:
            comment = self.comment
        textpieces.append(comment)

        return textpieces

    def _nice_print(self) -> str:
        text = """
parent: {}
MT channel: {}
daughter: {}
comment: {}
""".format(self.parent, self.MT, self.daughter, self.comment)
        return text

    @classmethod
    def from_text(cls, text: str) -> Reaction:
        """
        Create a Reaction object from text

        Parameters
        ----------
        cls : TYPE
            DESCRIPTION.
        text : str
            formatted text describing the reaction.

        Returns
        -------
        Reaction
            Reaction object.

        """
        # Split the reaction in its components
        pieces = PAT_SPACE.split(text.strip())
        parent = pieces[0].strip()
        MT = pieces[1]
        daughter = pieces[2]
        # the rest is comments
        comment = ''
        if len(pieces) > 3:
            for piece in pieces[3:]:
                comment = comment+' '+piece

        comment = comment.strip()

        return cls(parent, MT, daughter, comment=comment)

    def __repr__(self) -> str:
        return self._nice_print()

    def __str__(self) -> str:
        return self._nice_print()



"""
Parsing of MCNP output file (.o)

Only a few features are implemented for the moment being.
"""

"""
Copyright 2019 F4E | European Joint Undertaking for ITER and the Development of
Fusion Energy (‘Fusion for Energy’). Licensed under the EUPL, Version 1.2 or - 
as soon they will be approved by the European Commission - subsequent versions
of the EUPL (the “Licence”). You may not use this work except in compliance
with the Licence. You may obtain a copy of the Licence at: 
    https://eupl.eu/1.2/en/
Unless required by applicable law or agreed to in writing, software distributed
under the Licence is distributed on an “AS IS” basis, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the Licence permissions
and limitations under the Licence.
"""

import os
import re
import logging
import pandas as pd
import pyvista as pv
import numpy as np

from f4enix.constants import SCIENTIFIC_PAT, PAT_DIGIT
from f4enix.input.MCNPinput import Input

# -- Identifiers --
SURFACE_ID = 'currently being tracked has reached surface'
CELL_ID = 'other side of the surface from cell'
POINT_ID = 'x,y,z coordinates:'
# -- Patterns --
PAT_NPS_LINE = re.compile(r' source')
# patComments = re.compile(r'(?i)C\s+')
# patUniverse = re.compile(r'(?i)u=\d+')
# patNPS = re.compile(r'(?i)nps')
# patNPS_value = re.compile(r'\s+[eE.0-9]+\s*')
# patXYZ = re.compile(r'\s+[eE.0-9]+\s[eE.0-9]+\s[eE.0-9]+\s*')
# patSDEF = re.compile(r'(?i)sdef')
# patSDEFsur = re.compile(r'(?i)sur=\d+')

STAT_CHECKS_COLUMNS = ['TFC bin behaviour', 'mean behaviour',
                       'rel error value', 'rel error decrease',
                       'rel error decrease rate', 'VoV value', 'VoV decrease',
                       'VoV decrease rate', 'FoM value', 'FoM behaviour',
                       'PDF slope']


class Output:
    def __init__(self, filepath: os.PathLike) -> None:
        """Object representing and MCNP output file

        Parameters
        ----------
        filepath : os.PathLike
            path to the output file to be parsed

        Attributes
        ----------
        filepath: os.PathLike
            path to the original output file
        name: str
            name of the original file
        lines: list[str]
            list of all the lines of the file

        Examples
        --------
        >>> from f4enix.output.MCNPoutput import Output
        ... # parse the output
        ... outp = Output('test.o')
        ... # print excel and .vtk file containing info on lost particles
        ... outp.print_lp_debug('outfile_name')
        ... # Get the results of the statistical checks for a specific tally
        ... outp.get_tally_stat_checks(46)
                                mean behaviour    rel error      rel error  ...
                                                    value         decrease  ...
        TFC bin behaviour
        desired                   random           <0.10               yes  ...
        observed                decrease            0.03               yes  ...
        passed?                       no             yes               yes  ...

        This tables are produced only if the bins have non-zero values and
        refer to the total bin. These results, combined with a summary check
        on all tally bins are used to compile the total summary table

        >>> # Get the total summary table
        ... outp.get_stat_checks_table()
                    mean            rel error                        TFC
                    behaviour        value          ...             bins
        Cell
        2              yes           yes            ...            Passed
        4              NaN           NaN            ...         All zeros
        6              NaN           NaN            ...         All zeros
        12             yes           yes            ...            Passed
        14             NaN           NaN            ...         All zeros
        24             NaN           NaN            ...         All zeros
        34             NaN           NaN            ...         All zeros
        22             yes           yes            ...            Passed
        32             yes           yes            ...            Passed
        44              no           yes            ...            Missed
        46              no           yes            ...            Missed
        104             no           yes            ...            Missed

        Get an MCNP table from the output file in a pandas DataFrame format:

        >>> outp.get_table(60)
        		 cell	mat	  ... photon wt generation
        2	1.0	  1	     0	  ... 	-1.000E+00
        3	2.0	  2	     1	  ...	-1.000E+00
        4	3.0	  3	     0	  ...	-1.000E+00
        5	4.0	  4	     0	  ...	-1.000E+00
        """
        self.filepath = filepath
        self.name = os.path.basename(filepath).split('.')[0]
        logging.info('reading {} ...'.format(self.name))
        self.lines = self._read_lines()
        logging.info('reading completed')

    def _read_lines(self) -> list[str]:
        lines = []
        with open(self.filepath, 'r', errors="surrogateescape") as infile:
            for line in infile:
                lines.append(line)
        return lines

    def get_NPS(self, particle: str = 'neutron') -> int:
        """Get the number of particles simulated.

        Parameters
        ----------
        particle: str
            source particle to be looked for. The trigger for founding
        the correct table will be '{particle} creation'. The default is
        'neutron'.


        Returns
        -------
        int
            number of particles simulated
        """
        PAT_NPS_TRIGGER = re.compile(particle+' creation')
        for i, line in enumerate(self.lines):
            if PAT_NPS_TRIGGER.search(line) is not None:
                nps = int(PAT_DIGIT.search(self.lines[i+3]).group())
                logging.info('NPS found: {}'.format(nps))
                return nps

        raise ValueError('No NPS could be read from file')

    def print_lp_debug(self, outpath: os.PathLike, print_video: bool = False,
                       get_cosine: bool = True,
                       input_mcnp: Input = None) -> None:
        """prints both an excel ['LPdebug_{}.vtp'] and a vtk cloud point file
        ['LPdebug_{}.vtp'] containing information about the lost particles
        registered in an MCNP run. A .csv file is also printed with origin
        and flight direction of each lost particle.

        Parameters
        ----------
        outpath : os.PathLike
            path to the folder where outputs will be dumped.
        print_video : bool, optional
            if True print the LP to video. deafult is False
        get_cosine : bool, optional
            if True recover also the cosines of the flight direction of each
            lost particle. By default is True
        input_mcnp : Input, optional
            Input file that generated the MCNP output. Providing this will
            ensure that also the universe in which the particles are
            lost will be tracked. By default is None.
        """

        # -- Variables --
        surfaces = []
        cells = []
        universes = []
        x = []
        y = []
        z = []
        u = []
        v = []
        w = []

        logging.info(
            'Recovering lost particles surfaces and cells in '+self.name)

        for i, line in enumerate(self.lines):

            if line.find(SURFACE_ID) != -1:  # LP in surface
                surfaces.append(PAT_DIGIT.search(line).group())

                cell_ID = PAT_DIGIT.search(self.lines[i+1]).group()
                cells.append(cell_ID)
                # add the universe if requested
                if input_mcnp is not None:
                    universe = input_mcnp.cells[cell_ID].get_u()
                    universes.append(universe)

                point = SCIENTIFIC_PAT.findall(self.lines[i+6])  # [0:3]
                cosines = SCIENTIFIC_PAT.findall(self.lines[i+7])  # [0:3]
                x.append(float(point[0]))
                y.append(float(point[1]))
                z.append(float(point[2]))
                if get_cosine:
                    u.append(float(cosines[0]))
                    v.append(float(cosines[1]))
                    w.append(float(cosines[2]))

            # if line.find(CELL_ID) != -1 and input_mcnp is not None:
            #     # LP is in cell
            #     cells.append(PAT_DIGIT.search(line).group())

            # if line.find(POINT_ID) != -1:  # LP in cell
            #     point = SCIENTIFIC_PAT.findall(line)  # [0:3]
            #     x.append(float(point[0]))
            #     y.append(float(point[1]))
            #     z.append(float(point[2]))

            #     if '***' in self.lines[i-10]:
            #         gp = SCIENTIFIC_PAT.findall(self.lines[i-9])[0:3]
            #     else:
            #         gp = SCIENTIFIC_PAT.findall(self.lines[i-10])[0:3]
            #     try:
            #         gp[2]
            #         for i in range(len(gp)):
            #             gp[i] = gp[i][0:-3]+'E'+gp[i][-3:]
            #         gp = '   '.join(gp)
            #         globalpointList.append(gp)
            #     except:
            #         globalpointList.append('NO')

        # Don't do nothing if no particles are lost
        if len(surfaces) == 0:
            logging.info('No particles were lost, no dumps to be done')
            return

        # Building the df
        df = pd.DataFrame()
        df['Surface'] = surfaces
        df['Cell'] = cells
        df['x'] = x
        df['y'] = y
        df['z'] = z
        if get_cosine:
            df['u'] = u
            df['v'] = v
            df['w'] = w
        if len(universes) > 0:
            df['Universe'] = universes

        # Get a complete set of locations
        loc = df.drop_duplicates().set_index(['Surface', 'Cell']).sort_index()

        # Count multiple occasions
        df['count'] = 1
        global_df = df[['Cell', 'Surface', 'count']]
        global_df = global_df.groupby(['Cell', 'Surface']).sum()
        global_df = global_df.sort_values(by='count', ascending=False)
        logging.debug('global df was built')

        # --- Printing to excel ---
        logging.info('printing to excel')

        # dump in the excel output
        filename = 'LPdebug_{}.{}'
        outfile = os.path.join(outpath, filename.format(self.name, 'xlsx'))
        with pd.ExcelWriter(outfile) as writer:
            # global df
            global_df.to_excel(writer, sheet_name='global')
            # loc.to_excel(writer, sheet_name='Locations')
            if input_mcnp is not None:
                logging.info('building universe df')
                u_df = df[['Universe', 'Cell', 'Surface', 'count']]
                u_df = u_df.groupby(['Universe', 'Cell', 'Surface']).sum()
                u_df = u_df.sort_values(by='count', ascending=False)
                u_df.to_excel(writer, sheet_name='by universe')

        # Dump also a csv with only the data
        if get_cosine:
            cols = ['x', 'y', 'z', 'u', 'v', 'w']
        else:
            cols = ['x', 'y', 'z']
        lp = df[cols]
        outfile = os.path.join(outpath, filename.format(self.name, 'csv'))
        lp.to_csv(outfile, index=None)

        # visualize the cloud point
        logging.info('building the cloud point')
        point_cloud = pv.PolyData(loc[['x', 'y', 'z']].values)

        if print_video:
            point_cloud.plot()

        outfile = os.path.join(outpath, filename.format(self.name, 'vtp'))
        point_cloud.save(outfile)

        logging.info('dump completed')

    def get_statistical_checks_tfc_bins(self) -> dict[int, str]:
        """
        Retrieve the result of the 10 statistical checks for all tallies.
        They are registered as either 'Missed', 'Passed' or 'All zeros' in a
        dictionary indicized using the tallies numbers.

        Returns
        -------
        stat_checks : dict[int, str]
            keys are the tally numbers, values the result of the statistical
            checks.

        """
        # Some global key words and patterns
        start_stat_check = 'result of statistical checks'
        miss = 'missed'
        passed = 'passed'
        allzero = 'no nonzero'
        pat_tnumber = re.compile(r'\s*\t*\s*\d+')
        end = 'the 10 statistical checks are only'

        # Recover statistical checks
        statcheck_flag = False
        stat_checks = {}
        for line in self.lines:
            if line.find(start_stat_check) != -1:
                statcheck_flag = True

            elif statcheck_flag:
                # Control if is a tally line
                tallycheck = pat_tnumber.match(line)
                if tallycheck is not None:
                    tnumber = int(tallycheck.group())
                    if line.find(miss) != -1:
                        result = 'Missed'
                    elif line.find(passed) != -1:
                        result = 'Passed'
                    elif line.find(allzero) != -1:
                        result = 'All zeros'
                    else:
                        print('Warning: tally n.'+str(tnumber) +
                              ' not retrieved')

                    stat_checks[tnumber] = result

                elif line.find(end) != -1:
                    # Exit from loop when all checks are read
                    break

        return stat_checks

    def get_tally_stat_checks(self, cell: int) -> pd.DataFrame:
        """Get the table of statistical checks for a specific tally

        Parameters
        ----------
        cell : int
            index of the cell to be parsed

        Returns
        -------
        pd.DataFrame
            table reporting the results of the statistical checks

        Raises
        ------
        ValueError
            if the cell is not found in the file.
        """
        trigger = re.compile(
            '           results of 10 statistical .+\s{}\n'.format(cell))
        found = False

        for i, line in enumerate(self.lines):
            if trigger.match(line):
                # found trigger
                found = True
                break
        if found:
            skiprows = i+5
            nrows = 3
            df = pd.read_csv(self.filepath, skiprows=skiprows, nrows=nrows,
                             header=None, sep=r'\s+')
            df.columns = STAT_CHECKS_COLUMNS
            df.set_index('TFC bin behaviour', inplace=True)

            # values for the pdf slopes are assigned wrongly to FoM
            for row in ['observed', 'passed?']:
                if pd.isna(df.loc[row, 'PDF slope']):
                    # pass the values to the correct columns
                    val = df.loc[row, 'FoM value']
                    df.loc[row, 'FoM value'] = np.nan
                    df.loc[row, 'PDF slope'] = val

                    if row == 'passed?':
                        # assigned passed to the empty ones
                        df.loc[row, 'FoM value'] = 'yes'
                        df.loc[row, 'FoM behaviour'] = 'yes'

        else:
            raise ValueError('Cell {} was not found'.format(cell))

        return df

    def get_stat_checks_table(self) -> pd.DataFrame:
        """get a table summarizing the 10 statisical checks results for each
        tally.

        one row per tally, one column per statistical check. An extra column
        is added that report the results of the statistical checks in all other
        TFC bins.

        Returns
        -------
        pd.DataFrame
            summary of the statistical checks results.
        """
        # get all available cells from the summary
        summary = self.get_statistical_checks_tfc_bins()
        rows = []
        for cell in list(summary.keys()):
            new_row = [int(cell)]
            try:
                table = self.get_tally_stat_checks(int(cell))
                new_row.extend(list(table.loc['passed?'].values))
            except ValueError as e:
                # If all bins have zero values it is expected not to find it
                if summary[cell] == 'All zeros':
                    new_row.extend([np.nan]*10)
                else:
                    # If they are not all zeros raise the exception
                    raise e
            rows.append(new_row)

        df = pd.DataFrame(rows)
        columns = ['Cell']
        columns.extend(STAT_CHECKS_COLUMNS[1:])
        df.columns = columns
        df.set_index('Cell', inplace=True)
        df['Other TFC bins'] = pd.Series(summary)

        return df.sort_index()

    def get_table(self, table_num: int) -> pd.DataFrame:
        """Extract a printed table from the MCNP output file.

        All tables should be accessible from their MCNP index.

        Parameters
        ----------
        table_num : int
            MCNP table index

        Returns
        -------
        pd.DataFrame
            parsed table

        Raises
        ------
        ValueError
            If the table associated to the requested index is not found
        """
        pat_table = re.compile('table '+str(table_num))

        skip = None
        look_total = False
        nrows = None

        # look for the trigger
        for i, line in enumerate(self.lines):
            if look_total:
                if line.find('total') != -1:
                    # help inferring the columns width using the data,
                    # it is more reliable
                    infer_line = self.lines[i-2]
                    widths = self._get_fwf_format_from_string(infer_line)
                    nrows = i-skip-2
                    break

            if pat_table.search(line) is not None:
                skip = i+1
                look_total = True

        if skip is None or nrows is None:
            raise ValueError(
                'Table {} not found or does not exists'.format(table_num))

        df = pd.read_fwf(self.filepath, skiprows=skip, nrows=nrows,
                         widths=widths, header=None)
        # the first n rows will actually be the title of the columns.
        # Check the first column to understand where the data starts
        columns = []
        for i in range(10):
            if not pd.isna(df.iloc[i, 0]):
                break

            for j in range(df.shape[1]):
                current_val = df.iloc[i, j]
                try:
                    if pd.isna(current_val):
                        current_val = ''
                except TypeError:
                    current_val = ''

                if i == 0:
                    # append the first part of string
                    columns.append(current_val)
                else:
                    # concat to previous string
                    columns[j] = columns[j]+' '+current_val

        # delete the headers rows
        df = df.iloc[i:, :]
        # add proper column header
        df.columns = columns
        # delete empty columns if any
        df.dropna(axis=1, how='all', inplace=True)

        return df

    @staticmethod
    def _get_fwf_format_from_string(line: str) -> None:
        # This works only with data that has no space
        new_datum = False
        widths = []
        counter = 0

        for i, char in enumerate(line):

            if char == ' ' and new_datum is False:
                new_datum = True
                widths.append(counter)
                counter = 1
            else:
                counter = counter+1
                if char != ' ':
                    new_datum = False

        # At the end of the  cycle add last spec
        widths.append(counter)

        # exclude dummy at the beginning and \n bin created at end of line
        return widths[1:]



"""
Plotter classes and methods for neutronics outputs.

For the moment it allows only to plot slices of meshes.
"""

"""
Copyright 2019 F4E | European Joint Undertaking for ITER and the Development of
Fusion Energy (‘Fusion for Energy’). Licensed under the EUPL, Version 1.2 or - 
as soon they will be approved by the European Commission - subsequent versions
of the EUPL (the “Licence”). You may not use this work except in compliance
with the Licence. You may obtain a copy of the Licence at: 
    https://eupl.eu/1.2/en/  
Unless required by applicable law or agreed to in writing, software distributed
under the Licence is distributed on an “AS IS” basis, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the Licence permissions
and limitations under the Licence.
"""

import os
import pyvista as pv
import numpy as np
import logging
import docx

from docx.shared import Inches, Mm
from docx.enum.text import WD_ALIGN_PARAGRAPH
from docx.enum.section import WD_ORIENT
from PIL import Image, ImageOps
import io
from copy import deepcopy
from abc import ABC, abstractmethod
import matplotlib.pyplot as plt
from matplotlib.ticker import PercentFormatter


from pathlib import Path

from math import radians, degrees
from f4enix.constants import TID_CATEGORIES, TNF_CATEGORIES, SDDR_CATEGORIES

pv.set_plot_theme('document')


class MeshPlotter:

    def __init__(self, mesh: pv.PolyData, stl: pv.PolyData = None
                 ) -> None:
        """Object responsible for the plotting of meshes.

        Allows slicing in different orientation and plotting of slices.

        Parameters
        ----------
        mesh : pv.PolyData
            mesh to be sliced
        stl : pv.PolyData, optional
            stl to be superimposed in the slices, by default None

        Attributes
        ----------
        mesh : pv.PolyData
            mesh to be sliced
        stl : pv.PolyData, optional
            stl to be superimposed in the slices
        legend_args : dict
            contains the default parameters for the plot of the scalar bars.
            use print(self.legend_args) to check the default values. The entire
            list of parameters that can be used can be found in the
            `pv.Plotter.add_scalar_bar() method <https://docs.pyvista.org/version/stable/api/plotting/_autosummary/pyvista.Plotter.add_scalar_bar.html>`_

        Examples
        --------
        Load a mesh and stl file and slice them in different way

        >>> from f4enix.output.plotter import MeshPlotter
        ... # Initialize the plotter with the mesh and stl
        ... # scaling the mesh and stl accordingly to be in the same units
        ... plotter = MeshPlotter(global_mesh.scale(0.01, inplace=False),
        ...                       stl=stl.scale(0.001, inplace=False))
        ... # There are many default settings that can be modified
        ... plotter.legend_args['vertical'] = False
        ... # -- Different slicing methods can be used: --
        ... # toroidal slicing
        ... toroidal_slices = plotter.slice_toroidal(30)  # 30 deg. increment
        ... # vertical slicing on an axis and defininig number of divisions
        ... horizontal_slices = plotter.slice_on_axis('z', 3)
        ... # General slicing using origin and normals for each slice
        ... slice_params = [
                ['slice 1', 0, 0, -11.500, 0, 0, 1],
                ['slice 2', 9, 0, -8.425, 0.25, 0, 0.9]]
        ... general_slices = plotter.slice(slice_params)

        and then plot the meshes

        >>> # Plot the slices
        ... array_name = 'scalar1'  # name of the scalar to be plot from the mesh
        ... images = plotter.plot_slices(horizontal_slices, array_name,
        ...                              n_colors=7,  # colors of the legend
        ...                              min_max=(1e6, 1e12),  # limits
        ...                              scale_title='Gy/h')  # legend title

        """
        self.mesh = mesh

        self.stl = stl

        if stl is not None:
            self._has_stl = True
        else:
            self._has_stl = False

        # Default legend settings
        self.legend_args = dict(
            title_font_size=18,
            label_font_size=18,
            shadow=True,
            # n_labels=7,
            italic=True,
            fmt="%.e",
            font_family="arial",
            vertical=True,
            position_x=0.85,
            position_y=0.25)

    def _get_plotter(self) -> pv.Plotter:
        # Initiate the plotter with all default actions if needed
        pl = pv.Plotter(off_screen=True)
        # TODO horizontal bar may be issue here
        pl.add_axes(viewport=(0.8, 0, 1, 0.2))
        # pl.set_background('white')

        return pl

    def slice(self, slice_param: list[list]
              ) -> list[tuple[str, pv.PolyData, pv.PolyData | None]]:
        """Perform arbitrary general slicing providing origin and normal
        vectors.

        Parameters
        ----------
        slice_param : list[list]
            matrix of parameters to be provided, one row for each slice.
             [[name, x, y, z, normx, normy, normz], ...]

        Returns
        -------
        list[tuple[str, pv.PolyData, pv.PolyData | None]]
            contains a list of (slice name, mesh slice, stl slice).
            the names are assigned according to slicing logic.
            In case of no stl assigned to the plotter, the stl slice will be
            equal to None.

        """

        outp = []
        for params in slice_param:
            name = params[0]
            origin = params[1:4]
            norm = params[4:]

            mesh_slice = self.mesh.slice(origin=origin, normal=norm)

            # get the corresponding stl slices if needed
            if self._has_stl:
                stl_slice = self._get_stl_slices([mesh_slice])[0]
            else:
                stl_slice = None

            outp.append((name, mesh_slice, stl_slice))

        return outp

    def slice_on_axis(self, axis: str, n: int
                      ) -> list[tuple[str, pv.PolyData, pv.PolyData | None]]:
        """Creates a series of slice of the mesh distributed equally along the
        specified axis. Stl file will be sliced accordingly if present

        Parameters
        ----------
        axis : str
            either 'x', 'y' or 'z'
        n : int
            number of divisions

        Returns
        -------
        list[tuple[str, pv.PolyData, pv.PolyData | None]]
            contains a list of (slice name, mesh slice, stl slice).
            the names are assigned according to slicing logic.
            In case of no stl assigned to the plotter, the stl slice will be
            equal to None.

        """
        # Use the automatic slicing from pyvista
        idxs = {'x': 0, 'y': 1, 'z': 2}

        # perform the slices
        slices = self.mesh.slice_along_axis(n, axis=axis)
        # get the corresponding stl slices if needed
        if self._has_stl:
            stl_slices = self._get_stl_slices(slices)

        outp = []
        # build the output
        for i, mslice in enumerate(slices):
            # TODO maybe a smarter function here
            name = 'P{} = {}'.format(axis, round(mslice.center[idxs[axis]], 1))
            if self._has_stl:
                stl_slice = stl_slices[i]
            else:
                stl_slice = None

            outp.append((name, mslice, stl_slice))

        return outp

    def slice_toroidal(self, theta_increment: float,
                       center: list = [0, 0, 0],
                       min_max_theta: tuple[float, float] = None,
                       ) -> list[tuple[str, pv.PolyData, pv.PolyData | None]]:
        """Create a series of slices of the mesh distributed toroidally equally
        around the vertical axis (z). Stl file will be sliced accordingly if
        present

        Parameters
        ----------
        theta_increment : float
            theta increments to be performed for each slide
        center : list, optional
            x, y, z point to be used as center for the
            slicing, by default [0, 0, 0].
        min_max_theta : tuple(float, float), optional
            specify a minimum and max theta for the plots (degrees).
            This may help
            avoiding slices in empty regions in partial models if the mesh
            is not properly rotated. By default is None, meaning that the 
            slicing will start at theta = 0 and finish at theta = pi.

        Returns
        -------
        list[tuple[str, pv.PolyData, pv.PolyData | None]]
            contains a list of (slice name, mesh slice, stl slice).
            the names are assigned according to slicing logic.
            In case of no stl assigned to the plotter, the stl slice will be
            equal to None.
        """

        center = np.array(center)
        increment = radians(theta_increment)
        mesh_slices = []

        if min_max_theta is None:
            angles = np.arange(0, np.pi, increment)
        else:
            angles = np.arange(radians(min_max_theta[0]),
                               radians(min_max_theta[1]),
                               increment)

        for theta in angles:
            normal = np.array(
                [np.cos(theta), np.sin(theta), 0.0]).dot(np.pi / 2.0)

            mesh_slice = self.mesh.slice(origin=center, normal=normal)

            # try to access its normal, if it has it, it is not empty
            try:
                mesh_slice.cell_normals[0]
            except KeyError:
                # it means that nothing can be sliced here because the
                # the slice is empty
                logging.warning(
                    'No slice can be done at theta={} deg'.format(degrees(theta)))
                continue

            mesh_slices.append(mesh_slice)

        if self._has_stl:
            stl_slices = self._get_stl_slices(mesh_slices)

        outp = []
        # build the output
        for i, mslice in enumerate(mesh_slices):
            name = 'theta = {} deg'.format(round(degrees(angles[i]), 1))
            if self._has_stl:
                stl_slice = stl_slices[i]
            else:
                stl_slice = None

            outp.append((name, mslice, stl_slice))

        return outp

    def plot_slices(self,
                    slices: list[tuple[str, pv.PolyData, pv.PolyData | None]],
                    array_name: str,
                    outpath: os.PathLike = None,
                    min_max: tuple[float] = None,
                    log_scale: bool = True,
                    stl_color: str = 'white',
                    n_colors: int = 256,
                    scale_quality: float = 3,
                    scale_title: str = None,
                    custom_categories: str = None
                    ) -> list[tuple[str, Image.Image]]:
        """Plot a series of slices to an outpath folder. The slices names
        are used as file names.

        Parameters
        ----------
        slices : list[tuple[str, pv.PolyData, pv.PolyData  |  None]]
            list of slices to be plotted. Usually produced with MeshPlotter
            methods.
        array_name : str
            name of the scalar array to be plotted
        outpath : os.PathLike, optional
            path to the directory that will contain all the plots. This folder
            must be already created. Files with the same name will be
            overridden. default is None, meaning that images will not be saved
        min_max : tuple[float], optional
            min and max values to be set for the scalars in all plots
        log_scale : bool, optional
            if true, activate logarithmic scale, by default True
        stl_color : str, optional
            color for the stl, by default 'white'
        n_colors : int, optional
            number of discrete colors to be used in the legend, by default 256
        scale_quality : float, optional
            increase the resolution of the picture by a factor, by default 2
        scale_title : str, optional
            ovverride the title of the scalar legend. The defualt is None,
            meaning that the array name will be used.
        custom_categorical : str
            plot special categorical plots. Admitted categories are TID,
            TNF and SDDR.

        Returns
        -------
        list[tuple[str, Image.Image]]
            list of images of the slices

        Raises
        ------
        ValueError
            if the path do not exists
        """
        # Check if a categorical plot is requested
        if custom_categories is not None:
            if custom_categories == 'SDDR':
                ctg = SDDR_CATEGORIES
            elif custom_categories == 'TNF':
                ctg = TNF_CATEGORIES
            elif custom_categories == 'TID':
                ctg = TID_CATEGORIES
            else:
                raise ValueError(
                    '{} are not valid custom categories'.format(
                        custom_categories))

        # Check that outpath exists
        if outpath is not None:
            if not os.path.exists(outpath):
                raise ValueError('{} does not exists'.format(outpath))

        images = []

        scalar_bar_args = deepcopy(self.legend_args)
        if scale_title is not None:
            scalar_bar_args['title'] = scale_title

        # get a number of labels that pairs with the number of colors
        # works only with even numbers
        if n_colors < 30 and n_colors % 2 == 0:
            scalar_bar_args['n_labels'] = int(n_colors/2+1)

        for i, (name, mesh_slice, stl_slice) in enumerate(slices):

            pl = self._get_plotter()
            if custom_categories is not None:
                # Add category label to the mesh
                colors = self._add_categorization(
                    mesh_slice, array_name, ctg['values'], ctg['categories'],
                    ctg['colors'], name=custom_categories)
                scalars = custom_categories
                cmap = colors
                below_color = None
                above_color = None
                lscale = False
                categories = True
            else:
                scalars = array_name
                cmap = 'jet'
                below_color = 'grey'
                above_color = 'purple'
                lscale = log_scale
                categories = False

            pl.add_mesh(mesh_slice, scalars=scalars,
                        scalar_bar_args=scalar_bar_args,
                        log_scale=lscale,
                        below_color=below_color,
                        above_color=above_color,
                        clim=min_max, cmap=cmap,
                        n_colors=n_colors,
                        categories=categories)
            if stl_slice is not None:
                pl.add_mesh(stl_slice, color=stl_color)

            # if i == 0:
            #     # ensure that all pictures will have the same bounds
            #     bounds = np.array(mesh_slice.bounds)*0.8
            bounds = np.array(mesh_slice.bounds)

            self._set_perpendicular_camera(mesh_slice, pl, bounds=bounds)

            # Get the image from the plotter and trim it
            im = Image.fromarray(pl.screenshot(None, return_img=True))
            bbox = ImageOps.invert(im).getbbox()
            trimmed = im.crop(bbox)

            images.append((name, trimmed))

            if outpath is not None:
                filename = os.path.join(outpath, '{}.png'.format(name))
                pl.screenshot(filename, scale=scale_quality)

        return images

    @staticmethod
    def _set_perpendicular_camera(mesh_slice: pv.PolyData,
                                  pl: pv.Plotter,
                                  bounds: list = None) -> None:
        # align camera: focus on center, position at center + normal
        center = mesh_slice.center
        pl.camera.focal_point = center
        pl.camera.position = center - mesh_slice.cell_normals[0]
        # reset camera to put entire mesh in view
        if bounds is None:
            pl.reset_camera()
        else:
            # help put Y up in case of PZ plots
            if bounds[-2] == bounds[-1]:
                pl.set_viewup([0, 1, 0])
            pl.reset_camera(bounds=bounds)

    def _get_stl_slices(self, mesh_slices: list[pv.PolyData]
                        ) -> list[pv.PolyData] | None:
        stl_slices = []
        # get the correspondent stl_slices
        for mesh_slice in mesh_slices:
            # check it is not empty
            if len(mesh_slice[mesh_slice.array_names[0]]) == 0:
                stl_slices.append(None)
                continue

            norm = mesh_slice.cell_normals[0]
            stl_slice = self.stl.slice(normal=norm, origin=mesh_slice.center)
            if stl_slice.bounds is None:
                # This may happen if the stl is smaller than the mesh
                stl_slices.append(None)
                continue
            # give a very small translation in order to not be exactly
            # coincident, using the normal should be sufficient
            # but it needs to be properly scaled
            scale = np.abs(mesh_slice.points).mean()*1e-3
            stl_slice.translate(-norm*scale, inplace=True)
            stl_slices.append(stl_slice)

        return stl_slices

    @staticmethod
    def _add_categorization(mesh: pv.PolyData, array_name: str,
                            vals: list[float], categories: list[str],
                            colors: list[str], name='label') -> list[str]:

        values = mesh[array_name]
        labels = np.empty(len(values), dtype='<U10')

        # categories = list(range(0, len(vals)+1))
        labels[:] = categories[-1]

        vals.reverse()
        categories.reverse()
        for val, ctg in zip(vals, categories):
            labels[values < val] = ctg

        mesh[name] = labels

        colors_to_use = []
        categories.reverse()
        for category, color in zip(categories, colors):
            if (labels == category).sum() > 1:
                colors_to_use.append(color)

        return colors_to_use

    # @staticmethod
    # def _get_perpendicular_vector(v):
    #     if v[1] == 0 and v[2] == 0:
    #         if v[0] == 0:
    #             raise ValueError('zero vector')
    #         else:
    #             return np.cross(v, [0, 1, 0])

    #     return np.cross(v, [1, 0, 0])


class Atlas:

    def __init__(self, name: str = 'atlas',
                 landscape: str = True) -> None:
        """Class to handle the generation of the Atlas.

        Parameters
        ----------
        name : str, optional
            atlas name, by default 'atlas'
        landscape : str, optional
            if true the atlas will be produced in landscape orientation,
            by default True

        Attributes
        ----------
        name : str
            atlas name, by default 'atlas'
        doc : docx.Document
            word document used for the creation of the atlas
        default_width : float
            width used to rescale images that are added to the atlas.
            by default this is set to 0.9*text_length in the document.
            to change it, it is recommended to use either the
            docx.shared.Inches or docx.shared.Inches.Mm conversion method.

        Examples
        --------
        Build an atlas adding sections manually (recommended, since images
        do not need to be saved to the disk). The images can be produced using
        py:method:`f4enix.output.plotter.MeshPlotter.plot_slices`. Both a 
        Word and PDF version are saved.

        >>> from f4enix.output.plotter import Atlas
        ... atlas = Atlas('Some title for the atlas')
        ... # reduce for instance the default width for the images
        ... atlas.default_width = atlas.default_width*0.9
        ... # add a section to the atlas containing images produced with
        ... # plot_slices
        ... atlas.add_section('new section', images)
        ... atlas.save('path/to/outfolder')

        """

        self.name = name
        doc = docx.Document()
        doc.add_heading(name, level=0)
        self.doc = doc  # Word Document
        if landscape:
            self._change_orientation()

        # set default width
        section = self.doc.sections[-1]
        margin = 10  # mm
        section.top_margin = Mm(margin)
        section.bottom_margin = Mm(margin)
        section.left_margin = Mm(margin)
        section.right_margin = Mm(margin)
        width = (section.page_width -
                 section.left_margin -
                 section.right_margin)/36000*0.9  # mm
        self.default_width = Mm(width)

    def _insert_img(self, img: os.PathLike, width=None) -> None:
        if width is None:
            width = self.default_width
        self.doc.add_picture(img, width=width)
        # self.doc.add_picture(img, height=Inches(7.5))
        last_paragraph = self.doc.paragraphs[-1]
        last_paragraph.alignment = WD_ALIGN_PARAGRAPH.CENTER

    def add_section(self, section_name: str,
                    images: list[tuple[str, Image.Image]],
                    level: int = 1,
                    include_section_name: bool = True,
                    disclaimer: str = None) -> None:
        """Add a section of plots to the Atlas.

        add a chapter to atlas, the level can be decided.

        Parameters
        ----------
        section_name : str
            Name of the section
        images : list[tuple[str, Image.Image]]
            list of images produced by the plot_slices() method to be added
            to the atlas.
        level : int, optional
            nested level where the section has to be added, by default 1
        include_section_name : bool
            If True, the section name is added to the image name, by default
            True
        disclaimer: str
            add a disclaimer equal for all plots to their titles, by default is
            None

        """

        self.doc.add_heading(section_name, level=level)

        for name, image in images:
            if include_section_name:
                name = '{} - {}'.format(name, section_name)
            if disclaimer is not None:
                name = name + ' ' + disclaimer
            self.doc.add_heading(name, level=level+1)
            # Get a binary stream for pythondocx
            imdata = io.BytesIO()
            image.save(imdata, format='png')
            imdata.seek(0)
            self._insert_img(imdata)
            # Clean the buffer
            del imdata

    def build_from_root(self, root_path: os.PathLike) -> None:
        """Given a root folder, build an atlas with all the pictures found
        in the subfolders. Only one level of subfolders is supported at the
        moment.

        Parameters
        ----------
        root_path : os.PathLike
            path to the root folder.
        """

        # TODO only one level of folders admitted for the moment
        for folder in os.listdir(root_path):
            folderpath = os.path.join(root_path, folder)

            if not os.path.isdir(folderpath):
                continue

            self.doc.add_heading(folder, level=1)

            # This ensures correct order of the pictures that will be the same
            # of creation
            files = os.listdir(folderpath)
            paths = []
            for file in files:
                paths.append(os.path.join(folderpath, file))

            for filepath in sorted(paths, key=os.path.getmtime):
                heading = os.path.basename(filepath)[:-4]
                ext = os.path.basename(filepath)[-3:]
                if ext not in ['jpg', 'png', 'tif']:
                    continue

                self.doc.add_heading(heading, level=2)
                self._insert_img(filepath)

    def _change_orientation(self) -> docx.section.Section:
        current_section = self.doc.sections[-1]
        new_width, new_height = (current_section.page_height,
                                 current_section.page_width)
        new_section = self.doc.add_section()
        new_section.orientation = WD_ORIENT.LANDSCAPE
        new_section.page_width = new_width
        new_section.page_height = new_height

        return new_section

    def save(self, outpath: os.PathLike,
             # pdfprint: bool = True
             ) -> None:
        """
        Save word atlas and possibly export PDF

        Parameters
        ----------
        outpath : os.PathLike
            path to the folder where to save the atlas(es)

        Returns
        -------
        None.

        """
        outpath_word = os.path.join(outpath, self.name+'.docx')
        # outpath_pdf = os.path.join(outpath, self.name+'.pdf')

        try:
            self.doc.save(outpath_word)
        except FileNotFoundError as e:
            print(' The following is the original exception:')
            print(e)
            print('\n it may be due to invalid characters in the file name')

        # if pdfprint:
        #     in_file = outpath_word
        #     out_file = outpath_pdf

        #     try:
        #         word = win32com.client.Dispatch('Word.Application')
        #     except:
        #         raise NotImplementedError('Word not installed')

        #     try:
        #         doc = word.Documents.Open(in_file)
        #         doc.ExportAsFixedFormat(
        #             OutputFileName=out_file,
        #             # 17 = PDF output, 18=XPS output
        #             ExportFormat=17,
        #             OpenAfterExport=False,
        #             # 0=Print (higher res), 1=Screen (lower res)
        #             OptimizeFor=0,
        #             # 0=No bookmarks,
        #             # 1=Heading bookmarks only,
        #             # 2=bookmarks match word bookmarks
        #             CreateBookmarks=1,
        #             DocStructureTags=True)
        #     finally:
        #         try:
        #             doc.Close()
        #         except UnboundLocalError:
        #             word.Quit()
        #             raise RuntimeError(
        #                 'something went wrong in the PDF production')
        #         word.Quit()

    # @staticmethod
    # def _wrapper(paragraph, ptype):
    #     """
    #     Wrap a paragraph in order to add cross reference

    #     Parameters
    #     ----------
    #     paragraph : docx.Paragraph
    #         image to wrap.
    #     ptype : str
    #         type of paragraph to wrap

    #     Returns
    #     -------
    #     None.

    #     """
    #     if ptype == 'table':
    #         instruction = ' SEQ Table \\* ARABIC'
    #     elif ptype == 'figure':
    #         instruction = ' SEQ Figure \\* ARABIC'
    #     else:
    #         raise ValueError(ptype+' is not a supported paragraph type')

    #     run = run = paragraph.add_run()
    #     r = run._r
    #     fldChar = OxmlElement('w:fldChar')
    #     fldChar.set(qn('w:fldCharType'), 'begin')
    #     r.append(fldChar)
    #     instrText = OxmlElement('w:instrText')
    #     instrText.text = instruction
    #     r.append(instrText)
    #     fldChar = OxmlElement('w:fldChar')
    #     fldChar.set(qn('w:fldCharType'), 'end')
    #     r.append(fldChar)

    # @staticmethod
    # def _highlightCell(cell, color='FBD4B4'):
    #     shading_elm_1 = parse_xml(r'<w:shd {} w:fill="'.format(nsdecls('w')) +
    #                               color + r'"/>')
    #     cell._tc.get_or_add_tcPr().append(shading_elm_1)


class Plotter2D(ABC):

    def __init__(self, suptitle: str = None, xlabel: str = None,
                 ylabel: str = None) -> None:
        """Abstract class for the definition of 2D plots.

        Parameters
        ----------
        suptitle : str, optional
            title of the plot, by default None
        xlabel : str, optional
            X axis label, by default None
        ylabel : str, optional
            Y axis label, by default None

        Attributes
        ----------
        fig : matplotlib.pyplot.Figure
            matplotlib figure representation
        ax : matplotlib.pyplot.Axis
            matplotlib axes representation
        markers : list[str]
            markers used in the plots ['o', 's', 'D', '^', 'X', 'p', 'd', '*'].
        lines : list[str]
            linestyles used in the plots ['-', '--', '-.', ':']
        colors : list[str]
            colors used in the plots. They compose a color-blind friendly
            palette.

        """
        fig, ax = plt.subplots()

        # Some default actions
        ax.grid(alpha=0.6, which='both')
        if suptitle is not None:
            ax.set_title(suptitle)
        if ylabel is not None:
            ax.set_ylabel(ylabel)
        if xlabel is not None:
            ax.set_xlabel(xlabel)

        # store the default
        self.fig = fig
        self.ax = ax

        # May be improved in the future with additional markers and colors
        # plot decorators
        self.markers = ['o', 's', 'D', '^', 'X', 'p', 'd', '*']*50
        self.lines = ['-', '--', '-.', ':']*50
        # Color-blind friendly palette
        self.colors = ['#377eb8', '#ff7f00', '#4daf4a', '#f781bf', '#a65628',
                       '#984ea3', '#999999', '#e41a1c', '#dede00']*50
        logging.debug('plotter initialized')

    @abstractmethod
    def plot(self) -> None:
        pass

    def save(self, outpath: os.PathLike) -> None:
        self.fig.savefig(outpath, dpi=300, bbox_inches='tight')
        logging.info('plot has been saved at {}'.format(outpath))


class CDFplot(Plotter2D):
    def __init__(self, suptitle: str = None,
                 xlabel: str = None,
                 ylabel: str = None) -> None:
        """Plotter for cumulative distributions.

        all datasets of observations are automatically binnned and plotted as 
        (unfilled) histograms.

        Parameters
        ----------
        suptitle : str, optional
            title of the plot, by default None
        xlabel : str, optional
            X axis label, by default None
        ylabel : str, optional
            Y axis label, by default None

        Attributes
        ----------
        fig : matplotlib.pyplot.Figure
            matplotlib figure representation
        ax : matplotlib.pyplot.Axis
            matplotlib axes representation
        markers : list[str]
            markers used in the plots ['o', 's', 'D', '^', 'X', 'p', 'd', '*'].
        lines : list[str]
            linestyles used in the plots ['-', '--', '-.', ':']
        colors : list[str]
            colors used in the plots. They compose a color-blind friendly
            palette.

        """
        super().__init__(suptitle, xlabel, ylabel)

    def plot(self, values_list: list, bins: int = 10,
             datalabels: list[str] = None, perc: bool = True,
             outside_legend: bool = False, cut_y: float = None,
             cut_x: float = None) -> None:
        """plot the comulative distributions as discrete steps

        Parameters
        ----------
        values_list : list
            list of list of values containing the occurencies to be analyzed
        bins : int, optional
            number of bins into which the observations should be divided,
            by default 10
        datalabels : list[str], optional
            list of labels to be added in the plot legend representing the
            different datasets reported in values_list, by default None
        perc : bool, optional
            if True, format y-axis as percentages, by default True
        outside_legend : bool
            if True pushes the legend outside of the graph frame, by default is
            False
        cut_y : float, optional
            cut the y axis to a specific value, by default is None.
        cut_x : float, optional
            cut the x axis to a specific value, by default is None.
        """
        # check that the length of labels is consisting
        if datalabels is not None:
            try:
                assert len(datalabels) == len(values_list)
            except AssertionError:
                msg = 'leghth of values ({}) is different from lenght of labels ({})'
                raise ValueError(msg.format(len(values_list), len(datalabels)))

        for i, values in enumerate(values_list):
            if datalabels is not None:
                label = datalabels[i]
            else:
                label = None

            self.ax.hist(values, bins=bins, histtype='step',
                         weights=np.ones(len(values))/len(values),
                         cumulative=True, label=label, linestyle=self.lines[i],
                         color=self.colors[i])
            if perc:
                self.ax.yaxis.set_major_formatter(PercentFormatter(1))

        if outside_legend:
            self.ax.legend(framealpha=1, bbox_to_anchor=(1, 1))
        else:
            self.ax.legend(loc='lower right', framealpha=1)

        if cut_y is not None:
            self.ax.set_ylim(top=cut_y)
        if cut_x is not None:
            self.ax.set_xlim(right=cut_x)
        logging.info('CDF was plotted')
